using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Association_mining
{
    /// <summary>
    /// Class responsible for displaying single charts.
    /// </summary>
    public partial class Chart : Form
    {
        /// <summary>
        /// A variable specifying the chart type. 
        /// <list type="number">
        /// <item>1 signifies Confidence chart</item>
        /// <item>2 signifies Confirmation-S chart</item>
        /// <item>3 signifies Confirmation-F chart</item>
        /// <item>4 signifies Anti-support chart</item>
        /// </list>
        /// </summary>
        protected int CType;
        /// <summary>
        /// Number of "blue" (F>=0) rules.
        /// </summary>
        protected int BRCount;
        /// <summary>
        /// Number of "red" (F &lt; 0) rules.
        /// </summary>
        protected int RRCount;
        /// <summary>
        /// Minimal (or maximal) possible value on the Y axis. 
        /// </summary>
        protected double minYValue;
        /// <summary>
        /// Minimal (or maximal) possible value on the Y axis. defined by the minimal/maximal 
        /// value used to generate rules.
        /// </summary>
        protected double minYValueAbs;
        /// <summary>
        /// Minimal support for displayed rules.
        /// </summary>
        protected double minSup;
        /// <summary>
        /// Absolute minimal support defined by the minimal 
        /// support used to generate rules.
        /// </summary>
        protected double minSupAbs;
        /// <summary> 
        /// Graphics object representing the plane the chart is drawn on.
        /// </summary>
        protected Graphics CGraphics;
        /// <summary> 
        /// Last point the mouse cursor visited in selection mode. Used to draw the lasso.
        /// </summary>
        protected Point LastPoint = Point.Empty;
        /// <summary> 
        /// First point in drawing the lasso in selection mode. Defined by the mouse position
        /// during the MouseDown event.
        /// </summary>
        protected Point OrigPoint = Point.Empty;
        /// <summary> 
        /// Variable showing if the user is selecting rules. Used to tell the paint function to draw the lasso.
        /// </summary>
        protected bool selectMode = false;
        /// <summary> 
        /// A GraphicsPath object responsible for drawing the lasso.
        /// </summary>
        protected GraphicsPath Selection;
        /// <summary> 
        /// Painter object responsible for providing basic chart drawing functions.
        /// </summary>
        protected ChartUtil Painter;
        /// <summary> 
        /// A list of rules with the same conclusion to be drawn.
        /// </summary>
        protected List<Rule> Rules;

        /// <summary> 
        /// Default constructor.
        /// </summary>
        public Chart()
        {
            InitializeComponent();
        }

        /// <summary> 
        /// Creates a new chart object of a given 
        /// type based on the given rules and minimal support
        /// </summary>
        /// <param name="minsup">Absolute minimal support</param>
        /// <param name="ruleList">List of rules with the same conclusion</param>
        /// <param name="type">Defines the type of chart to draw
        /// <list type="number">
        /// <item>1 signifies Confidence chart</item>
        /// <item>2 signifies Confirmation-S chart</item>
        /// <item>3 signifies Confirmation-F chart</item>
        /// <item>4 signifies Anti-support chart</item>
        /// </list></param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="maxa">Maximal anti-support</param>
        public Chart(List<Rule> ruleList, int type, double minsup, double minconf,
                        double mins, double minf, double maxa)
        {
            InitializeComponent();
            Rules = new List<Rule>();
            Rules = ruleList;
            CType = type;
            minSupAbs = minSup = minsup;
            tboxSup.Text = minSupAbs.ToString();
            Painter = new ChartUtil();

            switch (type)
            {
                case 1:
                    this.Text = "Confidence (" + SetUtil.ParseHipothesis(ruleList[0]) + ")";
                    this.label2.Text = "Minimal confidence";
                    minYValueAbs = minYValue = minconf;
                    tboxConf.Text = minconf.ToString();
                    break;
                case 2:
                    this.Text = "Confirmation s (" + SetUtil.ParseHipothesis(ruleList[0]) + " )";
                    this.label2.Text = "Minimal confirmation s";
                    minYValueAbs = minYValue = mins;
                    tboxConf.Text = mins.ToString(); ;
                    break;
                case 3:
                    this.Text = "Confirmation f (" + SetUtil.ParseHipothesis(ruleList[0]) + ")";
                    this.label2.Text = "Minimal cofirmation f";
                    minYValueAbs = minYValue = minf;
                    tboxConf.Text = minf.ToString();
                    break;
                case 4:
                    this.Text = "Anti-support (" + SetUtil.ParseHipothesis(ruleList[0]) + ")";
                    this.label2.Text = "Maximum anti-support";
                    minYValueAbs = minYValue = maxa;
                    tboxConf.Text = maxa.ToString() ;
                    break;
                default:
                    this.Text = "Chart"; break;
            }
        }

        /// <summary> 
        /// Adjusts the window size to fit the drawing requirements.
        /// </summary>
        protected void AdjustParameters()
        {
            int reminder;
            reminder = pictureBox.Width % 12;
            pictureBox.Width -= reminder;
            Width -= reminder;
            reminder = pictureBox.Height % 12;
            pictureBox.Height -= reminder;
            Height -= reminder;
            BRCount = RRCount = 0;

            Double.TryParse(tboxConf.Text, out minYValue);
            Double.TryParse(tboxSup.Text, out minSup);

            minSup = Math.Min(minSup, 1);
            minSup = Math.Max(minSup, minSupAbs);
            minYValue = Math.Min(minYValue, 1);
            if(CType != 4)
                minYValue = Math.Max(minYValue, minYValueAbs);
            else
                minYValue = Math.Min(minYValue, minYValueAbs);

            switch (CType)
            {
                case 1: Painter.SetMinValues(minSup, minYValue, -1, -1, 1); break;
                case 2: Painter.SetMinValues(minSup, 0, minYValue, -1, 1); break;
                case 3: Painter.SetMinValues(minSup, 0, -1, minYValue, 1); break;
                case 4: Painter.SetMinValues(minSup, 0, -1, -1, minYValue); break;
            }
        }

        /// <summary> 
        /// Paints the chart.
        /// </summary>
        protected void PaintChart(object sender, PaintEventArgs e)
        {
            AdjustParameters();
            CGraphics = e.Graphics;
            CGraphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.Default;
            CGraphics.Clear(Color.White);
            Painter.DrawNet(CGraphics, pictureBox.Width, 
                            pictureBox.Height, CType, Rules[0], 8);
            CGraphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            if (cboxBorder.Checked)
            {
                Painter.DrawBorder(CGraphics, pictureBox.Width, pictureBox.Height,
                                   CType, Rules[0]);
            }

            if (cboxSelect.Checked)
            {
                Painter.DrawSelected(CGraphics, pictureBox.Width, pictureBox.Height,
                                    CType, Rules, ref RRCount, ref BRCount, 
                                    cboxInvisible.Checked, Selection, CType);
            }
            else if (cboxPareto.Checked && !cboxInvisible.Checked)
            {
                Painter.DrawParetoRule(CGraphics, pictureBox.Width, pictureBox.Height,
                                        CType, Rules, ref RRCount, ref BRCount);
            }
            else if (cboxPareto.Checked && cboxInvisible.Checked)
            {
                Painter.DrawParetoRuleVisible(CGraphics, pictureBox.Width, pictureBox.Height,
                                        CType, Rules, ref RRCount, ref BRCount);
            }
            else
            {
                Painter.DrawRule(CGraphics, pictureBox.Width, pictureBox.Height,
                                 CType, Rules,ref RRCount, ref BRCount, cboxInvisible.Checked);
            }

            if (selectMode)
                Painter.DrawLasso(CGraphics, Selection);

            Painter.DrawRuleCount(CGraphics, pictureBox.Width, pictureBox.Height,
                                  RRCount, BRCount, 8.25F);
        }

        /// <summary> 
        /// Forces the chart to repaint itself.
        /// </summary>
        protected void Change(object sender, EventArgs e)
        {
            pictureBox.Invalidate();
        }

        /// <summary> 
        /// Saves the chart in *.bmp format.
        /// </summary>
        protected void button1_Click(object sender, EventArgs e)
        {
            Bitmap bitmap = new Bitmap(pictureBox.Width, pictureBox.Height);
            pictureBox.DrawToBitmap(bitmap, pictureBox.Bounds);

            saveFileDialog1.FileName = this.Text.Replace('\\', '_').Replace('/', '_').
                           Replace(':', '_').Replace('*', '_').Replace('?', '_').
                           Replace('<', '_').Replace('>', '_').Replace('|', '_') + ".bmp";
            saveFileDialog1.Title = "Save as...";
            saveFileDialog1.Filter = "BMP - Windows Bitmap|*.bmp";
            saveFileDialog1.FilterIndex = 1;
            if (saveFileDialog1.ShowDialog() != DialogResult.Cancel)
                bitmap.Save(saveFileDialog1.FileName);
        }

        /// <summary> 
        /// Starts drawing the lasso.
        /// </summary>
        protected void Mouse_Down(object sender, MouseEventArgs e)
        {
            if (LastPoint != Point.Empty)
            {
                pictureBox.Invalidate();
                pictureBox.Update();
            }
            Selection = new GraphicsPath();
            LastPoint.X = e.X;
            LastPoint.Y = e.Y;
            OrigPoint = LastPoint;
            selectMode = true;
        }

        /// <summary> 
        /// Closes the lasso.
        /// </summary>
        protected void Mouse_Up(object sender, MouseEventArgs e)
        {
            int yValue = 0;
            int xValue, lastPareto = 0;
            bool paretoOptimal = true;
            bool clicked;
            bool visible = false;

            int CWidth = (int)pictureBox.Width;
            int CHeight = (int)pictureBox.Height;
            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            Selection.CloseFigure();
            OrigPoint = Point.Empty;
            if (Selection.GetBounds().Width < 2)
                selectMode = false;
            pictureBox.Invalidate();
            DGV.RowCount = 0;

            for (int idx = 0; idx < Rules.Count; idx++)
            {
                #region[pareto-optimality check]
                xValue = (int)((Rules[idx].Support * 10 + 1.0) * paddingW);
                switch (CType)
                {
                    case 1:
                        yValue = CHeight - (int)((Rules[idx].Confidence * 10.0 + 1.0) * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].Confidence >= minYValue);
                        if (idx == 0 || Rules[idx].Confidence > Rules[lastPareto].Confidence)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 2:
                        yValue = (CHeight - 6 * paddingH) - (int)(Rules[idx].S * 5.0 * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].S >= minYValue);
                        if (idx == 0 || Rules[idx].S > Rules[lastPareto].S)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 3:
                        yValue = (CHeight - 6 * paddingH) - (int)(Rules[idx].F * 5.0 * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].F >= minYValue);
                        if (idx == 0 || Rules[idx].F > Rules[lastPareto].F)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 4:
                        yValue = CHeight - (int)((Rules[idx].AntiSupport * 10.0 + 1.0) * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].AntiSupport <= minYValue);
                        if (idx == 0 || Rules[idx].AntiSupport < Rules[lastPareto].AntiSupport)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                }
                #endregion

                clicked = (e.X > xValue - 5 && e.X < xValue + 5 && 
                           e.Y > yValue - 5 && e.Y < yValue + 5);
                paretoOptimal = paretoOptimal || !cboxPareto.Checked;

                if ((Selection.IsVisible(xValue, yValue) || clicked) &&
                    ((visible && paretoOptimal) || cboxInvisible.Checked))
                {
                    AddToGrid(Rules[idx]);
                }            
            }
            DGV.AutoResizeColumns();
        }

        /// <summary> 
        /// Draws the lasso.
        /// </summary>
        protected void Mouse_Move(object sender, MouseEventArgs e)
        {
            if (OrigPoint != Point.Empty)
            {
                Point p = new Point(e.X, e.Y);
                Selection.AddLine(LastPoint, p);
                LastPoint.X = e.X;
                LastPoint.Y = e.Y;
                pictureBox.Invalidate();
                pictureBox.Update();
            }
        }

        /// <summary> 
        /// Adds a rule to the datagrid.
        /// </summary>
        /// <param name="r">Rule to be added to the datagrid</param>
        protected void AddToGrid(Rule r)
        {
            int n = DGV.RowCount++;
            DGV[0, n].Value = SetUtil.ParseEvidence(r); ;
            DGV[1, n].Value = SetUtil.ParseHipothesis(r);
            DGV[2, n].Value = String.Format("{0,2:F2}", r.Support);
            DGV[3, n].Value = String.Format("{0,2:F2}", r.Confidence);
            DGV[4, n].Value = String.Format("{0,2:F2}", r.S);
            DGV[5, n].Value = String.Format("{0,2:F2}", r.F);
            DGV[6, n].Value = String.Format("{0,2:F2}", r.AntiSupport);
        }
    }
}