using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Association_mining
{
    /// <summary> 
    /// Class responsible for providing chart drawing functions.
    /// </summary>
    public class ChartUtil
    {
        /// <summary> 
        /// Mnimal confidence.
        /// </summary>
        protected double minConf;
        /// <summary> 
        /// Mnimal confirmation-s
        /// </summary>
        protected double minS;
        /// <summary> 
        /// Mnimal confirmation-f.
        /// </summary>
        protected double minF;
        /// <summary> 
        /// Maximal anti-support.
        /// </summary>
        protected double maxAnti;
        /// <summary> 
        /// Mnimal support.
        /// </summary>
        protected double minSup;

        /// <summary> 
        /// Sets minimal values of all measures.
        /// </summary>
        /// <param name="mA">Maximal anti-support</param>
        /// <param name="mConf">Mnimal confidence</param>
        /// <param name="mF">Minimal confirmation-f</param>
        /// <param name="mS">Minimal confirmation-s</param>
        /// <param name="mSup">Minimal support</param>
        public void SetMinValues(double mSup, double mConf,
                                 double mS, double mF, double mA)
        {
            minConf = mConf;
            minS = mS; 
            minF = mF;
            maxAnti = mA;
            minSup = mSup;
        }

        /// <summary> 
        /// Draws a full (red) circle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        public void DrawFullCircle(Graphics g, int x, int y)
        {
            g.FillEllipse(new SolidBrush(Color.Red),
                        (float)x - 4, (float)y - 4, 9, 9);
        }

        /// <summary> 
        /// Draws a empty circle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        public void DrawEmptyCircle(Graphics g, int x, int y)
        {
            g.DrawEllipse(new Pen(Color.Black),
                        (float)x - 4, (float)y - 4, 9, 9);
        }

        /// <summary> 
        /// Draws a full (blue) triangle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        public void DrawFullTriangle(Graphics g, int x, int y)
        {
            Point ptC1 = new Point(x - 6, y + 3);
            Point ptC2 = new Point(x + 6, y + 3);
            Point ptC3 = new Point(x, y - 6);
            g.FillPolygon(new SolidBrush(Color.Blue), new Point[] { ptC1, ptC2, ptC3 });
        }

        /// <summary> 
        /// Draws a empty triangle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        public void DrawEmptyTriangle(Graphics g, int x, int y)
        {
            Point ptC1 = new Point(x - 6, y + 3);
            Point ptC2 = new Point(x + 6, y + 3);
            Point ptC3 = new Point(x, y - 6);
            g.DrawPolygon(new Pen(Color.Black), new Point[] { ptC1, ptC2, ptC3 });
        }

        /// <summary> 
        /// Draws a full (red) circle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        /// <param name="size">Size of the circle</param>
        public void DrawFullCircle(Graphics g, int x, int y, float size)
        {
            g.FillEllipse(new SolidBrush(Color.Red),
                        (float)x - size/2, (float)y - size/2, size, size);
        }

        /// <summary> 
        /// Draws a full (blue) triangle on Graphics g in the point x,y.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="x">X coordinate of the point to draw</param>
        /// <param name="y">Y coordinate of the point to draw</param>
        /// <param name="size">Size of the triangle</param>
        public void DrawFullTriangle(Graphics g, int x, int y, float size)
        {
            Point ptC1 = new Point((int)(x - 2 * size / 3), (int)(y + size / 3));
            Point ptC2 = new Point((int)(x + 2 * size / 3), (int)(y + size / 3));
            Point ptC3 = new Point(x, (int)(y - 2 * size / 3));
            g.FillPolygon(new SolidBrush(Color.Blue), new Point[] { ptC1, ptC2, ptC3 });
        }

        /// <summary> 
        /// Draws chart grid depending on the chart type.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="fontSize">Font size</param>
        /// <param name="r">Rule required to define confirmation-f borders</param>
        /// <param name="type">Chart type</param>
        public void DrawNet(Graphics g, int CWidth, int CHeight, 
                            int type, Rule r,  float fontSize )
        {
            string yText;
            float minX, minY, maxX, maxY;

            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            #region [type dependent options]
            switch (type)
            {
                #region[confirmation-s]
                case 2: 
                    yText = "confirmation s";
                    minX = 0.0F;
                    minY = -1.0F;
                    maxX = maxY = 1.0F;
                    g.FillRectangle(new SolidBrush(Color.LemonChiffon),
                                                new Rectangle(paddingW, paddingH,
                                                CWidth - 2 * paddingW,
                                                CHeight - 7 * paddingH));
                    g.FillRectangle(new SolidBrush(Color.Yellow),
                                                    new Rectangle(paddingW, 6 * paddingH,
                                                    CWidth - 2 * paddingW,
                                                    CHeight - 7 * paddingH));
                    //minimal support
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                    new Rectangle(paddingW, paddingH,
                                                    (int)(minSup * 10.0 * paddingW),
                                                    10 * paddingH));
                    //minimal confirmation-s
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                    new Rectangle(paddingW, 6 * paddingH - (int)(5 * minS * paddingH),
                                                    10 * paddingW,5 * paddingH + (int)(5 * minS * paddingH)));
                    break;
                #endregion
                #region[confirmation-f]
                case 3: 
                    yText = "confirmation f";
                    minX = 0.0F;
                    minY = -1.0F;
                    maxX = maxY = 1.0F;
                    g.FillRectangle(new SolidBrush(Color.LemonChiffon),
                                                    new Rectangle(paddingW, paddingH,
                                                    CWidth - 2 * paddingW,
                                                    CHeight - 7 * paddingH));
                    g.FillRectangle(new SolidBrush(Color.Yellow),
                                                    new Rectangle(paddingW, 6 * paddingH,
                                                    CWidth - 2 * paddingW,
                                                    CHeight - 7 * paddingH));
                    //minimal support
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                        new Rectangle(paddingW, paddingH,
                                                        (int)(minSup * 10.0 * paddingW),
                                                        10 * paddingH));
                    //minimal confirmation-f
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                            new Rectangle(paddingW, 6 * paddingH - (int)(5 * minF * paddingH),
                                                            10 * paddingW,
                                                            5 * paddingH + (int)(5 * minF * paddingH)));
                    /* gray region on confirmationf chart (on multichart) depending on
                     * thresholds for confidence and anti-support
                    //minimal confidence
                    double temp = ((r.a+r.b+r.c+r.d)*minConf - (r.a+r.b))/
                                  ((-r.a-r.b+r.c+r.d)*minConf + (r.a+r.b));

                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                            new Rectangle(paddingW, 6 * paddingH - (int)(5 * temp * paddingH),
                                                            10 * paddingW,
                                                            5 * paddingH + (int)(5 * temp * paddingH)));
                    //minimal anti-support
                    GraphicsPath gPath = new GraphicsPath();
                    for (double i = 0; i < 1 - maxAnti; i += 0.01)
                    {
                        double t1 = ((r.a + r.b + r.c + r.d) * i / (maxAnti + i) - (r.a + r.b)) /
                                  ((-r.a - r.b + r.c + r.d) * i / (maxAnti + i) + (r.a + r.b));
                        double t2 = ((r.a + r.b + r.c + r.d) * (i + 0.01) / (maxAnti + i + 0.01) - (r.a + r.b)) /
                                  ((-r.a - r.b + r.c + r.d) * (i + 0.01) / (maxAnti + i + 0.01) + (r.a + r.b));
                        gPath.AddLine(paddingW + (int)(10 * paddingW * i), 6 * paddingH - (int)(5 * paddingH * t1),
                            paddingW + (int)(10 * paddingW * (i + 0.01)), 6 * paddingH - (int)(5 * paddingH * t2));
                    }
                    double t3 = ((r.a + r.b + r.c + r.d) * 1 / (maxAnti + 1) - (r.a + r.b)) /
                                  ((-r.a - r.b + r.c + r.d) * 1 / (maxAnti + 1) + (r.a + r.b));
                    gPath.AddLine(paddingW + (int)(10 * paddingW * 1), 6 * paddingH - (int)(5 * paddingH * t3), paddingW + (int)(10 * paddingW * 1), CHeight - paddingH);
                    gPath.CloseFigure();
                    if (maxAnti > 0)
                        g.FillPath(new SolidBrush(Color.FromArgb(120, Color.Gray)), gPath);
                    else
                        g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)), paddingW, paddingH, 10 * paddingW, 10 * paddingH);
                     */ 
                    break;

                #endregion
                #region[anti-support]
                case 4: 
                    yText = "anti-support";
                    minX = minY = 0.0F;
                    maxX = maxY = 1.0F;
                    g.FillRectangle(new SolidBrush(Color.Yellow),
                                                new Rectangle(paddingW, paddingH,
                                                CWidth - 2 * paddingW,
                                                CHeight - 2 * paddingH));
                    g.FillPolygon(new SolidBrush(Color.LemonChiffon), new Point[] {
                                                    new Point(paddingW, CHeight - paddingH),
                                                    new Point(CWidth - paddingW,CHeight - (int)(paddingH*(10.0*((double)(r.a+r.b+r.c+r.d)/(double)(r.a+r.b)-1.0) + 1))),
                                                    new Point(CWidth - paddingW, CHeight - paddingH)});
                    //minimal support
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                    new Rectangle(paddingW, paddingH,
                                                    (int)(minSup * 10.0 * paddingW),
                                                    10 * paddingH));
                    //minimal anti-support
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                            new Rectangle(paddingW, paddingH,
                                                            10 * paddingW,
                                                            (int)(10.0 * (1.0 - maxAnti) * paddingH)));
                    /* gray region on anti-support chart (on multichart) depending on
                     * thresholds for confidence and confirmation f
                    
                    //minimal confidence
                    if(minConf>0)
                    g.FillPolygon(new SolidBrush(Color.FromArgb(120, Color.Gray)), new Point[] {
                                                    new Point(paddingW, CHeight - paddingH),
                                                    new Point(paddingW, paddingH),
                                                    new Point(CWidth - paddingW, paddingH),
                                                    new Point(CWidth - paddingW, CHeight - paddingH - (int)(10.0 * paddingH *(1/minConf - 1)))});
                    //minimal confirmation-f
                    double temp3 =
                        ((r.a + r.b) * minF + (r.a + r.b)) /
                        ((r.a + r.b + r.c + r.d) - minF * (-r.a - r.b + r.c + r.d));
                    if (temp3 > 0)
                        g.FillPolygon(new SolidBrush(Color.FromArgb(120, Color.Gray)), new Point[] {
                                                    new Point(paddingW, CHeight - paddingH),
                                                    new Point(paddingW, paddingH),
                                                    new Point(CWidth - paddingW, paddingH),
                                                    new Point(CWidth - paddingW, CHeight - paddingH - (int)(10.0 * paddingH *(1/temp3 - 1)))});
                    
                    g.FillRectangle(new SolidBrush(Color.White),
                                                    new Rectangle(0, 0,
                                                    CWidth - 1 * paddingW,
                                                    paddingH));
                     */ 
                    break;
#endregion
                #region[confidence]
                default: 
                    yText = "confidence";
                    minX = minY = 0.0F;
                    maxX = maxY = 1.0F;

                    g.FillRectangle(new SolidBrush(Color.LemonChiffon),
                                                new Rectangle(paddingW, paddingH,
                                                CWidth - 2 * paddingW,
                                                CHeight - 2 * paddingH));
                    g.FillRectangle(new SolidBrush(Color.Yellow),
                                                    new Rectangle(paddingW, CHeight - (int)(paddingH * (10.0 * ((double)(r.a + r.b) / (double)(r.a + r.b + r.c + r.d)) + 1)),
                                                    CWidth - 2 * paddingW,
                                                    (int)(paddingH * 10.0 * ((double)(r.a + r.b) / (double)(r.a + r.b + r.c + r.d)))));
                    //minimal support
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                        new Rectangle(paddingW, paddingH,
                                                        (int)(minSup * 10.0 * paddingW),
                                                        10 * paddingH));

                    //minimal confidence
                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                        new Rectangle(paddingW, CHeight - (int)((10.0 * minConf    + 1) * paddingH),
                                                        10 * paddingW,
                                                        (int)(10.0 * minConf * paddingH)));

                    /* gray region on confidence chart (on multichart) depending on
                     * thresholds for confirmation f and anti-support
                    
                    //minimal confirmation-f
                    double temp2 = 
                        ((r.a+r.b)*minF + (r.a+r.b))/
                        ((r.a+r.b+r.c+r.d) - minF*(-r.a-r.b+r.c+r.d));

                    g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)),
                                                        new Rectangle(paddingW, CHeight - (int)((10.0 * temp2 + 1) * paddingH),
                                                        10 * paddingW,
                                                        (int)(10.0 * temp2 * paddingH)));
                    //maximal anti-support
                    GraphicsPath gPath2 = new GraphicsPath();
                    for (double i = 0; i < 1-maxAnti; i += 0.01)
                    {
                        gPath2.AddLine(paddingW + (int)(10 * paddingW * i), 11 * paddingH - (int)(10 * paddingH * i / (maxAnti + i)),
                            paddingW + (int)(10 * paddingW * (i + 0.01)), 11 * paddingH - (int)(10 * paddingH * (i + 0.01) / (maxAnti + i + 0.01)));
                    }
                    gPath2.AddLine(paddingW+(int)(10*paddingW*1), 11*paddingH - (int)(10*paddingH* 1 / (maxAnti + 1)),paddingW+(int)(10*paddingW*1), CHeight-paddingH);
                    gPath2.CloseFigure();
                    if (maxAnti > 0)
                        g.FillPath(new SolidBrush(Color.FromArgb(120, Color.Gray)), gPath2);
                    else
                        g.FillRectangle(new SolidBrush(Color.FromArgb(120, Color.Gray)), paddingW, paddingH, 10 * paddingW, 10 * paddingH);
                     */ 
                    break;
                #endregion
            }
            #endregion

            Pen coloredPen = new Pen(Color.Black);
            Rectangle chartRect = new Rectangle(paddingW, paddingH,
                                      CWidth - 2 * paddingW,
                                      CHeight - 2 * paddingH);
            g.DrawRectangle(coloredPen, chartRect);
            coloredPen.DashStyle = DashStyle.Dash;

            double xStep = (Math.Abs(minX) + Math.Abs(maxX)) / 10.0;
            double yStep = (Math.Abs(minY) + Math.Abs(maxY)) / 10.0;

            for (int i = 0; i < 11; i++)
            {
                g.DrawLine(coloredPen, (i + 1) * paddingW, paddingH, (i + 1) * paddingW, CHeight - paddingH);
                g.DrawLine(coloredPen, paddingW, (i + 1) * paddingH, CWidth - paddingW, (i + 1) * paddingH);
                g.DrawString(String.Format("{0,4:F1}", maxY - i * yStep),
                    new Font("Lucida Console", fontSize), new SolidBrush(Color.Black), paddingW - 4*fontSize, (i + 1) * paddingH - fontSize);
                g.DrawString(String.Format("{0,4:F1}", minX + i * xStep, 1),
                    new Font("Lucida Console", fontSize), new SolidBrush(Color.Black), (i + 1) * paddingW - 2 * fontSize, CHeight - paddingH + fontSize - 2);
            }
            g.DrawString("support", new Font("Courier New", fontSize + 2, FontStyle.Bold), new SolidBrush(Color.Black), CWidth - paddingW - 6 * fontSize, 11*paddingH + (int)(2*fontSize));
            g.DrawString(yText, new Font("Courier New", fontSize + 2, FontStyle.Bold), new SolidBrush(Color.Black), paddingW, paddingH - (int)(2.5*fontSize));
        }

        /// <summary> 
        /// Draws rules.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="type">Chart type</param>
        /// <param name="brCount">Number of "blue" (F >= 0) rules</param>
        /// <param name="rrCount">Number of "red" (F &lt; 0) rules</param>
        /// <param name="invisible">Tells whether "invisible" rules should be drawn</param>
        /// <param name="ruleList">Rule list</param>
        public void DrawRule(Graphics g, int CWidth, int CHeight, 
                            int type, List<Rule> ruleList, ref int rrCount, 
                            ref int brCount, bool invisible)
        {
            int yValue = 0;
            int xValue;
            bool visible = false;

            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            foreach (Rule r in ruleList)
            {
                visible = (r.Support >= minSup && r.Confidence >= minConf &&
                           r.S >= minS && r.F >= minF && r.AntiSupport <= maxAnti);
                xValue = (int)((r.Support * 10 + 1.0) * paddingW);
                switch (type)
                {
                    case 1: yValue = CHeight - (int)((r.Confidence * 10.0 + 1.0) * paddingH);break;
                    case 2: yValue = (CHeight - 6 * paddingH) - (int)(r.S * 5.0 * paddingH);break;
                    case 3: yValue = (CHeight - 6 * paddingH) - (int)(r.F * 5.0 * paddingH);break;
                    case 4: yValue = CHeight - (int)((r.AntiSupport * 10.0 + 1.0) * paddingH);break;
                }

                if (visible)
                {
                    if (r.F <= 0)
                    {
                        DrawFullCircle(g, xValue, yValue);
                        rrCount += 1;
                    }
                    else
                    {
                        DrawFullTriangle(g, xValue, yValue);
                        brCount++;
                    }
                }
                else if (invisible)
                {
                    if (r.F <= 0)
                        DrawEmptyCircle(g, xValue, yValue);
                    else
                        DrawEmptyTriangle(g, xValue, yValue);
                }
            }
        }

        /// <summary> 
        /// Draws only the selected rules.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="type">Chart type</param>
        /// <param name="brCount">Number of "blue" (F >= 0) rules</param>
        /// <param name="rrCount">Number of "red" (F &lt; 0) rules</param>
        /// <param name="invisible">Tells whether "invisible" rules should be drawn</param>
        /// <param name="ruleList">Rule list</param>
        /// <param name="selection">Lasso</param>
        /// <param name="selectMode">Defines which chart (if it's a multichart) 
        /// has the selectin lasso on it</param>
        public void DrawSelected(Graphics g, int CWidth, int CHeight, 
                                int type, List<Rule> ruleList, ref int rrCount, 
                                ref int brCount, bool invisible, 
                                GraphicsPath selection, int selectMode)
        {
            int yValue = 0;
            int xValue, ySelectValue;
            bool visible = false;

            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;


            foreach (Rule r in ruleList)
            {
                visible = (r.Support >= minSup && r.Confidence >= minConf &&
                           r.S >= minS && r.F >= minF && r.AntiSupport <= maxAnti);
                xValue = (int)((r.Support * 10 + 1.0) * paddingW);
                switch (type)
                {
                    case 1: yValue = CHeight - (int)((r.Confidence * 10.0 + 1.0) * paddingH); break;
                    case 2: yValue = (CHeight - 6 * paddingH) - (int)(r.S * 5.0 * paddingH); break;
                    case 3: yValue = (CHeight - 6 * paddingH) - (int)(r.F * 5.0 * paddingH); break;
                    case 4: yValue = CHeight - (int)((r.AntiSupport * 10.0 + 1.0) * paddingH); break;
                }
                switch (selectMode)
                {
                    case 1: ySelectValue = CHeight - (int)((r.Confidence * 10.0 + 1.0) * paddingH); break;
                    case 2: ySelectValue = (CHeight - 6 * paddingH) - (int)(r.S * 5.0 * paddingH); break;
                    case 3: ySelectValue = (CHeight - 6 * paddingH) - (int)(r.F * 5.0 * paddingH); break;
                    default: ySelectValue = CHeight - (int)((r.AntiSupport * 10.0 + 1.0) * paddingH); break;
                }

                if (selection != null && selection.IsVisible(xValue, ySelectValue) && visible)
                {
                    if (r.F <= 0)
                    {
                        DrawFullCircle(g, xValue, yValue);
                        rrCount++;
                    }
                    else
                    {
                        DrawFullTriangle(g, xValue, yValue);
                        brCount++;
                    }
                }
                else if (invisible)
                {
                    if (r.F <= 0)
                        DrawEmptyCircle(g, xValue, yValue);
                    else
                        DrawEmptyTriangle(g, xValue, yValue);
                }
            }
        }

        /// <summary> 
        /// Draws pareto rules (only).
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="type">Chart type</param>
        /// <param name="brCount">Number of "blue" (F >= 0) rules</param>
        /// <param name="rrCount">Number of "red" (F &lt; 0) rules</param>
        /// <param name="ruleList">Rule list</param>
        public void DrawParetoRule(Graphics g, int CWidth, int CHeight, 
                                   int type, List<Rule> ruleList, 
                                   ref int rrCount,ref int brCount)
        {
            int yValue = 0;
            int xValue, lastPareto = 0;
            bool paretoOptimal = true;
            bool visible = false;

            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            for (int idx = 0; idx < ruleList.Count; idx += ruleList[idx].EqualSupportCount)
            {

                visible = (ruleList[idx].Support >= minSup && ruleList[idx].Confidence >= minConf &&
                           ruleList[idx].S >= minS && ruleList[idx].F >= minF && ruleList[idx].AntiSupport <= maxAnti);

                switch (type)
                {
                    case 1:
                        yValue = CHeight - (int)((ruleList[idx].Confidence * 10.0 + 1.0) * paddingH);
                        if (idx == 0 || ruleList[idx].Confidence > ruleList[lastPareto].Confidence)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 2:
                        yValue = (CHeight - 6 * paddingH) - (int)(ruleList[idx].S * 5.0 * paddingH);
                        if (idx == 0 || ruleList[idx].S > ruleList[lastPareto].S)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 3:
                        yValue = (CHeight - 6 * paddingH) - (int)(ruleList[idx].F * 5.0 * paddingH);
                        if (idx == 0 || ruleList[idx].F > ruleList[lastPareto].F)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 4:
                        yValue = CHeight - (int)((ruleList[idx].AntiSupport * 10.0 + 1.0) * paddingH);
                        if (idx == 0 || ruleList[idx].AntiSupport < ruleList[lastPareto].AntiSupport)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                }

                xValue = (int)((ruleList[idx].Support * 10 + 1.0) * paddingW);

                if (paretoOptimal && visible)
                {
                    if (ruleList[idx].F <= 0)
                    {
                        DrawFullCircle(g, xValue, yValue);
                        rrCount++;
                    }
                    else
                    {
                        DrawFullTriangle(g, xValue, yValue);
                        brCount++;
                    }
                }
            }
        }

        /// <summary> 
        /// Draws pareto rules (and "invisible" rules as well.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="type">Chart type</param>
        /// <param name="brCount">Number of "blue" (F >= 0) rules</param>
        /// <param name="rrCount">Number of "red" (F &lt; 0) rules</param>
        /// <param name="ruleList">Rule list</param>
        public void DrawParetoRuleVisible(Graphics g, int CWidth, int CHeight,
                                          int type, List<Rule> ruleList,
                                          ref int rrCount, ref int brCount)
        {
            int yValue = 0;
            int xValue, lastPareto = 0;
            bool paretoOptimal = true;
            bool visible = false;

            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            for (int idx = 0; idx < ruleList.Count; idx++)
            {
                visible = (ruleList[idx].Support >= minSup && ruleList[idx].Confidence >= minConf &&
                           ruleList[idx].S >= minS && ruleList[idx].F >= minF && ruleList[idx].AntiSupport <= maxAnti);
                switch (type)
                {
                    case 1:
                        yValue = CHeight - (int)((ruleList[idx].Confidence * 10.0 + 1.0) * paddingH);
                        if (idx == 0 || ruleList[idx].Confidence > ruleList[lastPareto].Confidence)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 2:
                        yValue = (CHeight - 6 * paddingH) - (int)(ruleList[idx].S * 5.0 * paddingH);
                        if (idx == 0 || ruleList[idx].S > ruleList[lastPareto].S)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 3:
                        yValue = (CHeight - 6 * paddingH) - (int)(ruleList[idx].F * 5.0 * paddingH);
                        if (idx == 0 || ruleList[idx].F > ruleList[lastPareto].F)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                    case 4:
                        yValue = CHeight - (int)((ruleList[idx].AntiSupport * 10.0 + 1.0) * paddingH);
                        if (idx == 0 || ruleList[idx].AntiSupport < ruleList[lastPareto].AntiSupport)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                }

                xValue = (int)((ruleList[idx].Support * 10 + 1.0) * paddingW);

                if (paretoOptimal && visible)
                {
                    if (ruleList[idx].F <= 0)
                    {
                        DrawFullCircle(g, xValue, yValue);
                        rrCount++;
                    }
                    else
                    {
                        DrawFullTriangle(g, xValue, yValue);
                        brCount++;
                    }
                }
                else
                {
                    if (ruleList[idx].F <= 0)
                        DrawEmptyCircle(g, xValue, yValue);
                    else
                        DrawEmptyTriangle(g, xValue, yValue);
                }
            }
        }

        /// <summary> 
        /// Draws border.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="type">Chart type</param>
        /// <param name="r">Rule required to draw the border</param>
        public void DrawBorder(Graphics g, int CWidth, int CHeight, int type, Rule r)
        {
            Pen pen = new Pen(Color.Black, 2);
            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            if(type == 2)
                g.DrawLine(pen, paddingW, 6 * paddingH,CWidth - paddingW, 6 * paddingH);
            else if(type == 3)
                g.DrawLine(pen, paddingW, 6 * paddingH,CWidth - paddingW, 6 * paddingH);
            else if(type == 4)
                g.DrawLine(pen, new Point(paddingW, CHeight - paddingH), 
                    new Point(CWidth - paddingW, CHeight - (int)(paddingH * (10.0 * ((double)(r.a + r.b + r.c + r.d) / (double)(r.a + r.b) - 1.0) + 1))));
            else
                g.DrawLine(pen, paddingW, 
                CHeight - (int)(paddingH * (10.0 * ((double)(r.a + r.b) / (double)(r.a + r.b + r.c + r.d)) + 1)),
                CWidth - paddingW,
                CHeight - (int)(paddingH * (10.0 * ((double)(r.a + r.b) / (double)(r.a + r.b + r.c + r.d)) + 1)));
        }

        /// <summary> 
        /// Draws the  number of each type of rules.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="CHeight">Chart height</param>
        /// <param name="CWidth">Chart width</param>
        /// <param name="brCount">Number of "blue" (F >= 0) rules</param>
        /// <param name="rrCount">Number of "red" (F &lt; 0) rules</param>
        /// <param name="fontSize">Font size</param>
        public void DrawRuleCount(Graphics g, int CWidth, int CHeight, int rrCount, 
                                  int brCount, float fontSize)
        {
            int paddingW = CWidth / 12;
            int paddingH = CHeight / 12;

            Font font = new Font("Arial", fontSize, FontStyle.Bold);
            SolidBrush brush = new SolidBrush(Color.Black);
            DrawFullTriangle(g, CWidth - 15*(int)fontSize, paddingH - (int)(1.3 * fontSize), fontSize);
            DrawFullCircle(g, CWidth - 9 * (int)fontSize, paddingH - (int)(1.5 * fontSize), fontSize);
            g.DrawString(brCount.ToString(), font, brush, CWidth - 15 * (int)fontSize + 10, paddingH - (int)(2.2 * fontSize));
            g.DrawString(rrCount.ToString(), font, brush, CWidth - 9 * (int)fontSize + 10,  paddingH - (int)(2.2 *fontSize));
        }

        /// <summary> 
        /// Draws a lasso.
        /// </summary>
        /// <param name="g">Graphics object to draw on</param>
        /// <param name="selection">Selection defining the lasso</param>
        public void DrawLasso(Graphics g, GraphicsPath selection)
        {
            Pen coloredPen = new Pen(Color.Black);
            coloredPen.DashStyle = DashStyle.Dash;
            g.DrawPath(coloredPen, selection);
        }

    }
}
