using System;
using System.Collections.Generic;
using System.Text;

namespace Association_mining
{
    /// <summary>
    /// Class responsible for creating databases and
    /// holding there logical structure.
    /// </summary>
    public class RecordHolder
    {
        private byte[] values;
        private int NumOfTypes;
        private string[] AttributeNames;
        private string[] AttributeTypes;
        /// <summary>
        /// Array of long representing the support of each 
        /// atribute (column) in the database.
        /// </summary>
        public long[]  AttributeCount;

        /// <summary>
        /// Array of attribute type names.
        /// </summary>
        public string[] Types
        {
            get { return AttributeTypes;  }
        }

        /// <summary>
        /// Array of distinct attribute type names.
        /// </summary>
        public string[] DistinctTypes
        {
            get 
            {
                List<string> distinct = new List<string>(0);
                for (int i = 0; i < AttributeTypes.Length; i++)
                {
                    if (!distinct.Contains(AttributeTypes[i]))
                        distinct.Add(AttributeTypes[i]);
                }
                return distinct.ToArray();
            }
        }

        /// <summary>
        /// Number of distinct types.
        /// </summary>
        public int TypeCount
        {
            get { return NumOfTypes; }
            set { NumOfTypes = value;}
        }

        /// <summary>
        /// Array of attribute names.
        /// </summary>
        public string[] Names
        {
            get { return AttributeNames; }
        }

        /// <summary>
        /// Number of names.
        /// </summary>
        public int NameCount
        {
            get { return AttributeNames.Length; }
        }

        /// <summary>
        /// Returns support for the specified attribute
        /// </summary>
        /// <param name="idx">Column index</param>
        /// <returns>The support of the attribte in column <c>idx</c> in the database</returns>
        public long AttributeSupport(int idx)
        {
            return AttributeCount[idx];
        }

        /// <summary>
        /// Constructor. Creates a new <c>REcordHolder</c> based on 
        /// the given set of types and names.
        /// </summary>
        /// <param name="AttrNamesDef">Array of attribute names</param>
        /// <param name="AttrTypesDef">Array of attribute types</param>
        public RecordHolder(string[] AttrNamesDef, string[] AttrTypesDef)
        {
            AttributeNames = new String[AttrNamesDef.Length];
            AttributeTypes = new String[AttrNamesDef.Length];
            AttributeCount = new long[AttrNamesDef.Length];
            values = new byte[AttrNamesDef.Length];

            for (int i = 0; i < AttrNamesDef.Length; i++)
            {
                AttributeTypes[i] = AttrTypesDef[i];
                AttributeNames[i] = AttrNamesDef[i];
            }
        }
    }

    /// <summary>
    /// Database class.
    /// </summary>
    public class Database
    {
        private List<byte[]> DynamicDataArray;
        private int RecordLength;
        private string Name;
        /// <summary>
        /// A template holding the logical structure of the database
        /// and giving a possibility to translate column meanings.
        /// </summary>
        public RecordHolder Template;
        /// <summary>
        /// A hash dictionary translating the concatenated attribute type and
        /// name into a database column number.
        /// </summary>
        public Dictionary<string, int> Translator = new Dictionary<string, int>();
        
        /// <summary>
        /// The number of all columns in the database. This number is equal to the
        /// number of distinct pairs Type-Name.
        /// </summary>
        public int BinaryColumnCount
        {
            get { return RecordLength; }
        }

        /// <summary>
        /// Number of distinct types in the database.
        /// </summary>
        public int ColumnCount
        {
            get { return Template.TypeCount; }
        }

        /// <summary>
        /// Number of records in the database.
        /// </summary>
        public int RowCount
        {
            get { return DynamicDataArray.Count; }
        }
        
        /// <summary>
        /// Database name.
        /// </summary>
        public string Text
        {
            get { return Name; }
            set { Name = value;  }
        }

        /// <summary>
        /// Default constructor.
        /// </summary>
        public Database() { }
        
        /// <summary>
        /// Constructor. Creates a new database on the base of the given 
        /// template.
        /// </summary>
        /// <param name="rHolder">Database template</param>
        /// <param name="name">Database name</param>
        public Database(RecordHolder rHolder, string name)
        {
            Template = rHolder;
            RecordLength = Template.NameCount;
            for (int i = 0; i < Template.NameCount; i++)
                Translator.Add(Template.Types[i]+Template.Names[i], i);
            DynamicDataArray = new List<byte[]>();
            Name = "Database "+name;
        }

        /// <summary>
        /// Allows to access any cell of the database.
        /// </summary>
        /// <param name="x">Database row</param>
        /// <param name="y">Database column</param>
        /// <returns>The value contained in the x,y database cell</returns>
        public byte this[int x, int y]
        {
            get { return DynamicDataArray[x][y];}
        }

        /// <summary>
        /// Adds a record to the database.
        /// </summary>
        /// <param name="record">Record</param>
        /// <param name="rlength">Record length</param>
        public void AddRecord(string[] record, int rlength)
        {
            byte[] values = new byte[RecordLength];
            int column;
            bool multiColumn = Template.DistinctTypes.Length > 1;

            for (int i = 0; i < rlength; i++)
            {
                if(multiColumn)
                    column = Translator[Template.DistinctTypes[i] + record[i]];
                else
                    column = Translator[Template.DistinctTypes[0] + record[i]];
                values[column] = 1;
                Template.AttributeCount[column]++;
            }
            DynamicDataArray.Add(values);
        }
    }
}
