using System;
using System.Collections.Generic;
using System.Text;
using Iesi.Collections;
using NGenerics.DataStructures.Trees;

namespace Association_mining
{
    /// <summary>
    /// Allows to compare support values in ore=der to sort them
    /// in descending order.
    /// </summary>
    public class SupportComparer : IComparer<long>
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public int Compare(long x, long y)
        {
            if (x >= y)
                return -1;
            else
                return 1;
        }
    }

    /// <summary>
    /// Class representing an node of an FPTree.
    /// </summary>
    public class FPNode
    {
        private Item m_item;
        private long m_sup;
        private int m_colID;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public FPNode() { }

        /// <summary>
        /// Creates a new node.
        /// </summary>
        /// <param name="t">Attribute type</param>
        /// <param name="n">Attribute name</param>
        /// <param name="id">Attribute id (database column)</param>
        /// <param name="sup">Node support</param>
        public FPNode(string t, string n, int id, long sup)
        {
            m_item = new Item(t,n);
            m_sup = sup;
            m_colID = id;
        }

        /// <summary>
        /// Item in the node.
        /// </summary>
        public Item NodeItem
        {
            get{return m_item;}
        }

        /// <summary>
        /// Node support.
        /// </summary>
        public long Support
        {
            get{return m_sup;}
            set{m_sup = value;}
        }

        /// <summary>
        /// Node item id.
        /// </summary>
        public int Id
        {
            get { return m_colID; }
        }
    }

    /// <summary>
    /// A header structure allowing to quickly localize all th nodes
    /// with the same item.
    /// </summary>
    public class Header
    {
        /// <summary>
        /// Item id.
        /// </summary>
        public int Id;
        
        /// <summary>
        /// Item support (not node support).
        /// </summary>
        public long Supp;
       
        /// <summary>
        /// List of items with the same item.
        /// </summary>
        public List<GeneralTree<FPNode>> ItemList;

        /// <summary>
        /// Creates a new header for the given item.
        /// </summary>
        /// <param name="id">Item id</param>
        /// <param name="supp">Item support</param>
        public Header(int id, long supp)
        {
            Id = id;
            Supp = supp;
            ItemList = new List<GeneralTree<FPNode>>();
        }
    }

    /// <summary>
    /// The FPTree class.
    /// </summary>
    public class FPTree
    {
        private GeneralTree<FPNode> m_tree;
        private Dictionary<int, int> colToHeader;
        private Header[] m_header;
        private GeneralTree<FPNode> m_currentNode;
        private IList<GeneralTree<FPNode>> m_childList;
        private bool m_singlePath;
        private bool m_isEmpty;

        /// <summary>
        /// Constructor. Plants a new FPTree with its header list, based on the
        /// given list of item ids sorted by descending support.
        /// </summary>
        /// <param name="sCols">List of item ids (database columns) 
        /// sorted by descending support</param>
        public FPTree(SortedDictionary<long, int> sCols)
        {
            m_tree = new GeneralTree<FPNode>(null);
            m_header = new Header[sCols.Count];
            colToHeader = new Dictionary<int, int>();

            int i=0;
            foreach (KeyValuePair<long, int> kvp in sCols)
            {
                colToHeader.Add(kvp.Value, i);
                m_header[i] = new Header(kvp.Value, kvp.Key);
                i++;
            }
            m_currentNode = m_tree;
            m_isEmpty = true;
            m_singlePath = true;
        }

        /// <summary>
        /// Tells if the tree consists only of a single path.
        /// </summary>
        public bool IsSinglePath
        {
            get {return m_singlePath;}
        }

        /// <summary>
        /// Tells if the tree has any nodes.
        /// </summary>
        public bool IsEmpty
        {
            get { return m_isEmpty; }
        }

        /// <summary>
        /// The header list of the tree.
        /// </summary>
        public Header[] ItemHeader
        {
            get { return m_header; }
        }

        /// <summary>
        /// Adds a new node to the FPTree. If from the current node being visited
        /// there is a child node with the same item then that child is becoming the 
        /// currently visited node and its support increased by the "added" node support.
        /// If there are no child nodes matching the newly added then a new one is 
        /// created with the support given.
        /// </summary>
        /// <param name="type">Item type</param>
        /// <param name="name">Item name</param>
        /// <param name="col">Item id (database column)</param>
        /// <param name="sup">Node support</param>
        public void Add(string type, string name, int col, long sup)
        {
                m_childList = m_currentNode.ChildNodes;
                int childCount = m_childList.Count;
                bool success = false;
                m_isEmpty = false;

                for (int child = 0; child < childCount; child++)
                {
                    //a child node representing the item already exists
                    if (m_childList[child].Data.Id == col)
                    {
                        m_childList[child].Data.Support+= sup;
                        m_currentNode = m_childList[child];
                        success = true;
                        break;
                    }
                }
                //no child represents the item - create new node
                if (!success)
                {
                    m_currentNode.Add(new GeneralTree<FPNode>(
                        new FPNode(type, name, col, sup)));
                    m_currentNode = m_currentNode.ChildNodes[childCount];
                    m_header[colToHeader[col]].ItemList.Add(m_currentNode);
                    if (childCount > 0)
                        m_singlePath = false;
                }
        }

        /// <summary>
        /// Sets the tree root as the currently visited node.
        /// </summary>
        public void  GoToRoot()
        {
            m_currentNode = m_tree;
        }

        /// <summary>
        /// Returns a list of nodes in the single path of the tree.
        /// </summary>
        /// <returns></returns>
        public Set SinglePathNodes()
        {
            GeneralTree<FPNode> child;
            Set path = new HashedSet(new FPNode[] {});
            GoToRoot();
            while (m_currentNode.ChildNodes.Count == 1)
            {
                child = m_currentNode.GetChild(0);
                path.Add(child.Data);
                m_currentNode = child;
            }
            return path;
        }
    }

    /// <summary>
    /// Static class providing functions needed to perform association mining
    /// with the Frequent Pattern Growth algorithm.
    /// </summary>
    public static class FPGrowth
    {
        private static List<FrequentSet> FrequentItems;
        private static int UniverseCount;
        private static double MinimalSupport; 

        /// <summary>
        /// Creates an FPTree based on a given database and performs the 
        /// FPGrowth algorithm.
        /// </summary>
        /// <param name="MainForm">Main program form</param>
        /// <param name="data">Database</param>
        /// <param name="ConclusionList">List of distinct conclusions</param>
        /// <param name="minsup">Minimal support</param>
        /// <param name="maxanti">Maximal anti-support</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <returns>A list of association rules generated with 
        /// the FPGrowth algorithm</returns>
        public static SortableBindingList<Rule> MiningAlgorithm(frmMain MainForm, Database data,
            List<Set> ConclusionList, double minsup, double minconf, double mins, double minf, 
            double maxanti, string decClass)
        {
            List<FrequentSet> fsList = new List<FrequentSet>();
            fsList = GenerateFrequentSets(data, minsup);
            if (decClass == "")
            {
                // Then generate rules from that list
                return SetUtil.GenerateRules(MainForm, data, ConclusionList, fsList, minconf, mins, minf, maxanti);
            }
            else
            {
                // Then generate rules from that list
                return SetUtil.GenerateRules(MainForm, data, ConclusionList, fsList, minconf, mins, minf, maxanti, decClass);
            }
        }

        /// <summary>
        /// Generates frequent sets using the FPGrowth algorithm.
        /// </summary>
        /// <param name="data">Database</param>
        /// <param name="minsup">Minimal support</param>
        /// <returns>Frequent sets with support greater or equal to the 
        /// given minimal support</returns>
        public static List<FrequentSet> GenerateFrequentSets(Database data, double minsup)
        {
            UniverseCount = data.RowCount;
            MinimalSupport = (double)(minsup * UniverseCount);
            FrequentItems = new List<FrequentSet>();
            //key - support, value - column number
            SortedDictionary<long, int> SortedCols = new SortedDictionary<long, int>(new SupportComparer());
            
            for (int i = 0; i < data.Template.NameCount; i++)
            {
                if (data.Template.AttributeSupport(i) >= MinimalSupport)
                    SortedCols.Add(data.Template.AttributeSupport(i),i);
            }

            #region[FPTree generation process]
            FPTree Tree = new FPTree(SortedCols);
            for (int row = 0; row < UniverseCount; row++)
            {
                Tree.GoToRoot();
                foreach(KeyValuePair<long,int> kvp in SortedCols)
                {
                    int col = kvp.Value;
                    if (data[row, col] == 1)
                    {
                       Tree.Add(data.Template.Types[col], 
                                data.Template.Names[col], col , 1);
                    }
                }
            }
            #endregion
            if(!Tree.IsEmpty)
                Grow(Tree, null);

            return FrequentItems;
        }

        /// <summary>
        /// The recursive Grow algorithm.
        /// </summary>
        /// <param name="tree">FPtree</param>
        /// <param name="alpha">Suffix set</param>
        public static void Grow(FPTree tree, FrequentSet alpha)
        {
            #region[single path in tree]
            if (tree.IsSinglePath)
            {
                List<Set> subsetList = SetUtil.Subsets(tree.SinglePathNodes());
                foreach (Set s in subsetList)
                {
                    Set combination = new HashedSet(new Item[] {});
                    long minSup = long.MaxValue;
                    foreach(FPNode fpn in s)
                    {
                        combination.Add(new Item(fpn.NodeItem.Type, fpn.NodeItem.Name));
                        if(fpn.Support < minSup) 
                            minSup = fpn.Support;
                    }
                    if (alpha != null)
                    {
                        foreach (FPNode fpn in alpha.set)
                            combination.Add(new Item(fpn.NodeItem.Type, fpn.NodeItem.Name));
                    }
                    if (combination.Count > 1 && minSup >= MinimalSupport)
                        FrequentItems.Add(new FrequentSet(combination,minSup,UniverseCount));
                }
            }
            #endregion
            //several paths to explore
            else
            {
                for (int i = tree.ItemHeader.Length - 1; i >= 0; i--)
                {
                    List<IList<GeneralTree<FPNode>>> prefixPaths =  new List<IList<GeneralTree<FPNode>>>();
                    SortedDictionary<long, int> SortedCols = new SortedDictionary<long, int>(new SupportComparer());
                    Dictionary<int, long> AttrSupp = new Dictionary<int, long>();
                    FrequentSet beta;

                    #region[Conditional pattern base generation]
                    foreach (GeneralTree<FPNode> bottomNode in tree.ItemHeader[i].ItemList)
                    {
                        long suffixSupport = bottomNode.Data.Support;
                        List<GeneralTree<FPNode>> path = new List<GeneralTree<FPNode>>();
                        foreach (GeneralTree<FPNode> node in bottomNode.Ancestors)
                        {
                            if (node.Data != null)
                                path.Add(new GeneralTree<FPNode>(
                                new FPNode(node.Data.NodeItem.Type, node.Data.NodeItem.Name,
                                        node.Data.Id, node.Data.Support)));
                        }  
                        foreach (GeneralTree<FPNode> node in path)
                        {  
                            node.Data.Support = suffixSupport;
                            if (AttrSupp.ContainsKey(node.Data.Id))
                                AttrSupp[node.Data.Id] += suffixSupport;
                            else
                                AttrSupp.Add(node.Data.Id, suffixSupport);
                        }
                        prefixPaths.Add(path);
                    }

                    foreach (KeyValuePair<int, long> kvp in AttrSupp)
                    {
                        if (kvp.Value >= MinimalSupport)
                            SortedCols.Add(kvp.Value,kvp.Key);
                    }
                    #endregion

                    #region[Conditional FPTree generation]
                    FPTree newTree = new FPTree(SortedCols);

                    foreach (List<GeneralTree<FPNode>> path in prefixPaths)
                    {
                        newTree.GoToRoot();
                        foreach (KeyValuePair<long, int> kvp in SortedCols)
                        {
                            foreach (GeneralTree<FPNode> node in path)
                            {
                                if (node.Data.Id == kvp.Value)
                                {
                                    newTree.Add(node.Data.NodeItem.Type,
                                             node.Data.NodeItem.Name,
                                             node.Data.Id, node.Data.Support);
                                }
                            }
                        }
                    }
                    #endregion

                    Set s = new HashedSet(new FPNode[]{});
                    s.Add(new FPNode(tree.ItemHeader[i].ItemList[0].Data.NodeItem.Type,
                                     tree.ItemHeader[i].ItemList[0].Data.NodeItem.Name,
                                     tree.ItemHeader[i].ItemList[0].Data.Id,
                                     tree.ItemHeader[i].Supp));
                    if (alpha != null)
                    {
                        foreach (FPNode fpn in alpha.set)
                        {
                            if (fpn.Support >= tree.ItemHeader[i].Supp)
                                s.Add(fpn);
                        }
                    }
                    beta = new FrequentSet(s, tree.ItemHeader[i].Supp, UniverseCount);
                    if (beta.Count > 1)
                    {
                        Set bs = new HashedSet(new Item[] { });
                        foreach (FPNode fpn in beta.set)
                            bs.Add(new Item(fpn.NodeItem.Type, fpn.NodeItem.Name));
                        FrequentItems.Add(new FrequentSet(bs, beta.Support, UniverseCount));
                    }

                    if(!newTree.IsEmpty)
                        Grow(newTree, beta);
                }
            }
        }

    }
}
