using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Collections;
using Iesi.Collections;

namespace Association_mining
{
    /// <summary>
    /// Contains functions responsible for reading and saving data and rule files.
    /// </summary>
    public static class FileProcessor
    {
        /// <summary>
        /// Opens an ARFF file.
        /// </summary>
        /// <param name="file">File</param>
        /// <returns>A database filled with records read from the file.
        /// Records that are incomplete are not taken into account.</returns>
        public static Database OpenARFFFile(FileStream file)
        {
            Database database = new Database();
            String line;
            Match match;

            String relationName="", AttrType;
            int AttrCount = 0;
            int TypeCount = 0;
            String[] Record;
            List<String> AttrTypeDef = new List<String>();
            List<String> AttrNameDef = new List<String>();
            Dictionary<int, bool> Banned = new Dictionary<int, bool>();
            int BannedCount = 0;

            RecordHolder RHolder;
            StreamReader sr = new StreamReader(file);

            while ((line= sr.ReadLine()) != null)
            {
                if ((match = Regex.Match(line, "@relation")).Success)
                {
                    relationName = "";
                    for (int i = match.Index + 10; i < line.Length; i++)
                        relationName = String.Concat(relationName, line[i]);
                }
                else if (Regex.Match(line, "@attribute").Success)
                {
                    AttrType = "";
                    for (int i = match.Index + 11; line[i] != ' '; i++)
                        AttrType = String.Concat(AttrType, line[i]);
                    //typ wyliczeinowy
                    if ((match = Regex.Match(line, @"{.*}")).Success)
                    {
                        string namelist = match.ToString();
                        namelist = namelist.Substring(1, namelist.Length - 2);
                        string[] names = Regex.Split(namelist, @"[ ]*,[ ]*");
                        foreach (string n in names)
                        {
                            AttrTypeDef.Add(AttrType);
                            AttrNameDef.Add(n);
                        }
                        TypeCount++;
                        Banned.Add(AttrCount, false);
                    }
                    //typ inny niz wyliczeniowy
                    else
                    {
                        Banned.Add(AttrCount,true);
                        BannedCount++;
                    }
                    AttrCount++;
                }
                else if (Regex.Match(line, "@data").Success)
                {
                    bool delete = false;
                    RHolder = new RecordHolder(AttrNameDef.ToArray(), AttrTypeDef.ToArray());
                    RHolder.TypeCount = TypeCount;
                    database = new Database(RHolder,relationName);
                    Regex expression = new Regex(@",[ ]*");
                    
                    Record = new String[AttrCount-BannedCount];
                    string[] attributes;
                    int recordCount;

                    while ((line = sr.ReadLine()) != "" && line != null)
                    { 
                        recordCount = 0;
                        attributes = expression.Split(line);

                        for (int i = 0; i < AttrCount; i++)
                        {
                            if (Banned[i])
                                continue;
                            if (attributes[i] == "?")
                                delete = true;
                            else
                            {
                                Record[recordCount] = attributes[i];
                                recordCount++;
                            }
                        }
                        if (!delete)
                            database.AddRecord(Record, recordCount);
                        else
                            delete = false;
                    }
                }
            }
            sr.Close();
            return database;
        }

        /// <summary>
        /// Saves a rule file
        /// </summary>
        /// <param name="file">File</param>
        /// <param name="ruleStructure">Rule structure</param>
        /// <param name="conclusionCount">Number of distinct conclusions in the structure</param>
        /// <param name="minsupp">Minimal support used to generate the rules</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="maxa">Maximal anti-support</param>
        public static void SaveRules(FileStream file, List<List<Rule>> ruleStructure, int conclusionCount, 
                                     double minsupp, double minconf, double mins, double minf, double maxa)
        {
            StreamWriter sw = new StreamWriter(file);

            sw.WriteLine(minsupp);
            sw.WriteLine(minconf);
            sw.WriteLine(mins);
            sw.WriteLine(minf);
            sw.WriteLine(maxa);
            sw.WriteLine(conclusionCount);

            foreach (List<Rule> ruleList in ruleStructure)
            {
                foreach (Rule r in ruleList)
                {
                    String helper = "";

                    helper += "{";
                    foreach (Item i in r.E)
                        helper += (i.Type + ", " + i.Name + ", ");
                    helper = helper.Substring(0, helper.Length - 2);
                    helper += "} {";

                    foreach (Item i in r.H)
                        helper += (i.Type + ", " + i.Name + ", ");
                    helper = helper.Substring(0, helper.Length - 2);
                    helper += "} {";

                    helper += r.a + ", ";
                    helper += r.b + ", ";
                    helper += r.c + ", ";
                    helper += r.d + ", ";
                    helper += r.EqualSupportCount + "}";
                    sw.WriteLine(helper);
                }
            }
            sw.Close();
        }

        /// <summary>
        /// Opens a rule file.
        /// </summary>
        /// <param name="file">File</param>
        /// <param name="Conclusions">Conclusion list</param>
        /// <param name="rules">Rule structure</param>
        /// <param name="minsupp">Minimal support that was used to generate the rules</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="maxa">Maximal anti-support</param>
        /// <returns></returns>
        public static List<List<Rule>> OpenRuleFile(FileStream file, List<Set> Conclusions, SortableBindingList<Rule> rules,
             out double minsupp, out double minconf, out double mins, out double minf, out double maxa)
        {
            StreamReader sr = new StreamReader(file);
            String line;
            Match match;
            Rule r;
            int idx = -1;
            int conclusionCount;
            Regex expression = new Regex(@"[ ]*,[ ]*");
            Regex lineRegex = new Regex(@"\{[^\{\}]*\}");

            minsupp = double.Parse(sr.ReadLine());
            minconf = double.Parse(sr.ReadLine());
            mins = double.Parse(sr.ReadLine());
            minf = double.Parse(sr.ReadLine());
            maxa = double.Parse(sr.ReadLine());
            conclusionCount = int.Parse(sr.ReadLine());

            List<List<Rule>> ruleStructure = new List<List<Rule>>(conclusionCount);
            for (int i = 0; i < conclusionCount; i++)
                ruleStructure.Add(new List<Rule>());

            while ((line= sr.ReadLine()) != null)
            {
                Set e = new HashedSet(new Item[] { });
                Set h = new HashedSet(new Item[] { });

                if ((match = lineRegex.Match(line)).Success)
                {
                    string evidenceList = match.ToString();
                    evidenceList = evidenceList.Substring(1, evidenceList.Length - 2);
                    string[] evidence = Regex.Split(evidenceList, @"[ ]*,[ ]*");
                    for (int i = 0; i < evidence.Length; i+=2 )
                        e.Add(new Item(evidence[i], evidence[i+1]));
                    
                    match  = match.NextMatch();
                    string hipothesisList = match.ToString();
                    hipothesisList = hipothesisList.Substring(1, hipothesisList.Length - 2);
                    string[] hipothesis = expression.Split(hipothesisList);
                    for (int i = 0; i < hipothesis.Length; i += 2)
                        h.Add(new Item(hipothesis[i], hipothesis[i + 1]));

                    match = match.NextMatch();
                    string parameterList = match.ToString();
                    parameterList = parameterList.Substring(1, parameterList.Length - 2);
                    string[] parameters = expression.Split(parameterList);

                    r = new Rule(e,h, long.Parse(parameters[0]),
                                            long.Parse(parameters[1]),
                                            long.Parse(parameters[2]),
                                            long.Parse(parameters[3]),
                                            int.Parse(parameters[4]));
                    rules.Add(r);
                    if (!Conclusions.Contains(h))
                    {
                        Conclusions.Add(h);
                        idx++;
                        ruleStructure[idx].Add(r);
                    }
                    else
                    {
                        ruleStructure[idx].Add(r);
                    }

                }
            }
            sr.Close();
            return ruleStructure ;
        }
        
        /// <summary>
        /// Opens a DATA file.
        /// </summary>
        /// <param name="file">File</param>
        /// <returns>A database filled with records read from the file.
        /// Records that are incomplete are not taken into account.</returns>
        public static Database OpenDATAFile(FileStream file)
        {
            
            Database database = new Database();
            String line;

            String relationName=file.Name;
            int AttrCount = 0;
            int TypeCount = 1;
            List<String> AttrTypeDef = new List<String>();
            List<String> AttrNameDef = new List<String>();
            Dictionary<string,string> NumToName = new Dictionary<string,string>();
            string[] Record = new string[AttrCount];
            RecordHolder RHolder;
            int i = 0;

            StreamReader sr = new StreamReader(file);

            while ((line= sr.ReadLine()) != null)
            {
                string[] attributes = Regex.Split(line, @"[ ]*,[ ]*");

                if (attributes[0] == "A")
                {
                    AttrTypeDef.Add("Site");
                    attributes[3] = attributes[3].Substring(1, attributes[3].Length - 2);
                    if(NumToName.ContainsValue(attributes[3]))
                        attributes[3] += "_"+attributes[1];
                    AttrNameDef.Add(attributes[3]);
                    NumToName.Add(attributes[1], attributes[3]);
                    AttrCount++;
                }
                else if (attributes[0] == "C")
                {
                    RHolder = new RecordHolder(AttrNameDef.ToArray(), AttrTypeDef.ToArray());
                    RHolder.TypeCount = TypeCount;
                    database = new Database(RHolder,relationName);
                    Record = new string[AttrCount];

                    while ((line = sr.ReadLine()) != "" && line != null)
                    {
                        if (line.StartsWith("V"))
                        {
                            Record[i] = NumToName[line.Substring(2,4)];
                            i++;
                        }
                        else
                        {
                            database.AddRecord(Record, i);
                            i = 0;
                            Record = new string[AttrCount];
                        }
                    }
                }
            }
            database.AddRecord(Record,i);
            sr.Close();
            return database;
        }

        /// <summary>
        /// Exports rules to Excel format.
        /// </summary>
        /// <param name="source">DataGridView displaying the rules</param>
        /// <param name="file">File</param>
        public static void ExportToExcel(DataGridView source, FileStream file)
        {
            StreamWriter excelDoc;

            excelDoc = new StreamWriter(file);

            const string startExcelXML = "<xml version>\r\n<Workbook " +
                  "xmlns=\"urn:schemas-microsoft-com:office:spreadsheet\"\r\n" +
                  " xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\n " +
                  "xmlns:x=\"urn:schemas-    microsoft-com:office:" +
                  "excel\"\r\n xmlns:ss=\"urn:schemas-microsoft-com:" +
                  "office:spreadsheet\">\r\n <Styles>\r\n " +
                  "<Style ss:ID=\"Default\" ss:Name=\"Normal\">\r\n " +
                  "<Alignment ss:Vertical=\"Bottom\"/>\r\n <Borders/>" +
                  "\r\n <Font/>\r\n <Interior/>\r\n <NumberFormat/>" +
                  "\r\n <Protection/>\r\n </Style>\r\n " +
                  "<Style ss:ID=\"BoldColumn\">\r\n <Font " +
                  "x:Family=\"Swiss\" ss:Bold=\"1\"/>\r\n </Style>\r\n " +
                  "<Style     ss:ID=\"StringLiteral\">\r\n <NumberFormat" +
                  " ss:Format=\"@\"/>\r\n </Style>\r\n <Style " +
                  "ss:ID=\"Decimal\">\r\n <NumberFormat " +
                  "ss:Format=\"0.0000\"/>\r\n </Style>\r\n " +
                  "<Style ss:ID=\"Integer\">\r\n <NumberFormat " +
                  "ss:Format=\"0\"/>\r\n </Style>\r\n <Style " +
                  "ss:ID=\"DateLiteral\">\r\n <NumberFormat " +
                  "ss:Format=\"mm/dd/yyyy;@\"/>\r\n </Style>\r\n " +
                  "</Styles>\r\n ";
            const string endExcelXML = "</Workbook>";

            int rowCount = 0;
            int sheetCount = 1;

            excelDoc.Write(startExcelXML);
            excelDoc.Write("<Worksheet ss:Name=\"Sheet" + sheetCount + "\">");
            excelDoc.Write("");
            excelDoc.Write("<Table>");
            excelDoc.Write("");
            excelDoc.Write("<Row>");
            for (int x = 0; x < source.Columns.Count; x++)
            {
                excelDoc.Write("<Cell ss:StyleID=\"BoldColumn\"><Data ss:Type=\"String\">");
                excelDoc.Write(source.Columns[x].Name);
                excelDoc.Write("</Data></Cell>");
            }
            excelDoc.Write("</Row>");
            foreach (DataGridViewRow x in source.Rows)
            {
                rowCount++;
                //if the number of rows is > 64000 create a new page to continue output
                if (rowCount == 64000)
                {
                    rowCount = 0;
                    sheetCount++;
                    excelDoc.Write("</Table>");
                    excelDoc.Write(" </Worksheet>");
                    excelDoc.Write("<Worksheet ss:Name=\"Sheet" + sheetCount + "\">");
                    excelDoc.Write("<Table>");
                }
                excelDoc.Write("<Row>"); //ID=" + rowCount + "
                for (int y = 0; y < source.Columns.Count; y++)
                {
                    System.Type rowType;
                    rowType = x.Cells[y].ValueType;
                    switch (rowType.ToString())
                    {
                        case "System.String":
                            string XMLstring = x.Cells[y].Value.ToString();
                            XMLstring = XMLstring.Trim();
                            XMLstring = XMLstring.Replace("&", "&amp;");
                            XMLstring = XMLstring.Replace(">", "&gt;");
                            XMLstring = XMLstring.Replace("<", "&lt;");
                            excelDoc.Write("<Cell ss:StyleID=\"StringLiteral\">" +
                                           "<Data ss:Type=\"String\">");
                            excelDoc.Write(XMLstring);
                            excelDoc.Write("</Data></Cell>");
                            break;
                        case "System.Boolean":
                            excelDoc.Write("<Cell ss:StyleID=\"StringLiteral\">" +
                                        "<Data ss:Type=\"String\">");
                            excelDoc.Write(x.Cells[y].Value.ToString());
                            excelDoc.Write("</Data></Cell>");
                            break;
                        case "System.Int16":
                        case "System.Int32":
                        case "System.Int64":
                        case "System.Byte":
                            excelDoc.Write("<Cell ss:StyleID=\"Integer\">" +
                                    "<Data ss:Type=\"Number\">");
                            excelDoc.Write(x.Cells[y].Value.ToString());
                            excelDoc.Write("</Data></Cell>");
                            break;
                        case "System.Decimal":
                        case "System.Double":
                            excelDoc.Write("<Cell ss:StyleID=\"Decimal\">" +
                                  "<Data ss:Type=\"Number\">");
                            excelDoc.Write(x.Cells[y].Value.ToString().Replace(',','.'));
                            excelDoc.Write("</Data></Cell>");
                            break;
                        case "System.DBNull":
                            excelDoc.Write("<Cell ss:StyleID=\"StringLiteral\">" +
                                  "<Data ss:Type=\"String\">");
                            excelDoc.Write("");
                            excelDoc.Write("</Data></Cell>");
                            break;
                        default:
                            throw (new Exception(rowType.ToString() + " not handled."));
                    }
                }
                excelDoc.Write("</Row>");
            }
            excelDoc.Write("</Table>");
            excelDoc.Write(" </Worksheet>");
            excelDoc.Write(endExcelXML);
            excelDoc.Close();
        }

        /// <summary>
        /// Exports rules to Excel format.
        /// </summary>
        /// <param name="source">DataGridView displaying the rules</param>
        /// <param name="file">File</param>
        public static void ExportToCSV(DataGridView source, FileStream file)
        {
            StreamWriter csvDoc;
            csvDoc = new StreamWriter(file);
            
            csvDoc.WriteLine("premise;conclusion;supp;conf;s;f;a-supp;");

            foreach (DataGridViewRow x in source.Rows)
            {
                for (int y = 0; y < source.Columns.Count; y++)
                {
                    System.Type rowType;
                    rowType = x.Cells[y].ValueType;
                    switch (rowType.ToString())
                    {
                        case "System.String":
                            csvDoc.Write(x.Cells[y].Value.ToString());
                            break;
                        case "System.Boolean":
                            csvDoc.Write(x.Cells[y].Value.ToString());
                            break;
                        case "System.Int16":
                        case "System.Int32":
                        case "System.Int64":
                        case "System.Byte":
                            csvDoc.Write(x.Cells[y].Value.ToString());
                            break;
                        case "System.Decimal":
                        case "System.Double":
                            csvDoc.Write(x.Cells[y].Value.ToString());
                            break;
                        case "System.DBNull":
                            csvDoc.Write("");
                            break;
                        default:
                            throw (new Exception(rowType.ToString() + " not handled."));
                    }
                    csvDoc.Write(";");
                }
                csvDoc.WriteLine("");
            }
            csvDoc.Close();
        }
    }

}
