using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Association_mining
{
    /// <summary>
    /// The multi chart class.
    /// </summary>
    public partial class MultiChart : Form
    {
        private int BRCount, RRCount;
        private double minConf, minS, minF, maxAnti, minSup;

        private double minSupAbs, minConfAbs, minSAbs, minFAbs, maxAntiAbs;
        private Point LastPoint = Point.Empty;
        private Point OrigPoint = Point.Empty;
        private int selectMode = 0;
        private List<Rule> Rules;
        private ChartUtil Painter;
        private GraphicsPath Selection;
        private int CHeight = 300;
        private int paddingW = 300 / 12;
        private int paddingH = 300 / 12;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public MultiChart()
        {
            InitializeComponent();
        }

        /// <summary> 
        /// Creates a new multichart object based on the given rules and minimal support
        /// </summary>
        /// <param name="minsup">Absolute minimal support</param>
        /// <param name="ruleList">List of rules with the same conclusion</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="maxa">Maximal anti-support</param>
        public MultiChart(List<Rule> ruleList, double minsup, double minconf,
                        double mins, double minf, double maxa)
        {
            InitializeComponent();
            Rules = new List<Rule>();
            Rules = ruleList;
            
            minSupAbs = minSup = minsup;
            minConfAbs = minConf = minconf;
            minSAbs = minS = mins;
            minFAbs = minF = minf;
            maxAntiAbs = maxAnti = maxa;

            tboxSup.Text = minSupAbs.ToString();
            tboxConf.Text = minconf.ToString();
            tboxS.Text = mins.ToString();
            tboxF.Text = minf.ToString();
            tboxAnti.Text = maxa.ToString();
            Painter = new ChartUtil();
            this.Text = "Multichart (" + SetUtil.ParseHipothesis(ruleList[0]) + ")";
        }

        /// <summary>
        /// Refreshes the multichart.
        /// </summary>
        protected void PaintCharts()
        {
            pictureBox1.Refresh();
            pictureBox2.Refresh();
            pictureBox3.Refresh();
            pictureBox4.Refresh();
        }
        
        /// <summary> 
        /// Saves the chart in *.bmp format.
        /// </summary>
        protected void button1_Click(object sender, EventArgs e)
        {
            Bitmap bitmap = new Bitmap(600, 600);
            pictureBox1.DrawToBitmap(bitmap, new Rectangle(0,0,300,300));
            pictureBox2.DrawToBitmap(bitmap, new Rectangle(300, 0, 300, 300));
            pictureBox3.DrawToBitmap(bitmap, new Rectangle(0, 300, 300, 300));
            pictureBox4.DrawToBitmap(bitmap, new Rectangle(300, 300, 300, 300));
            saveFileDialog1.FileName = this.Text.Replace('\\','_').Replace('/','_').
                                       Replace(':','_').Replace('*','_').Replace('?','_').
                                       Replace('<','_').Replace('>','_').Replace('|','_') + ".bmp";
            saveFileDialog1.Title = "Save as...";
            saveFileDialog1.Filter = "BMP - Windows Bitmap|*.bmp";
            saveFileDialog1.FilterIndex = 1;
            if (saveFileDialog1.ShowDialog() != DialogResult.Cancel)
                bitmap.Save(saveFileDialog1.FileName);
        }

        /// <summary> 
        /// Paints the multichart.
        /// </summary>
        protected void CalculateMinValues()
        {
            Double.TryParse(tboxConf.Text, out minConf);
            Double.TryParse(tboxSup.Text, out minSup);
            Double.TryParse(tboxS.Text, out minS);
            Double.TryParse(tboxF.Text, out minF);
            Double.TryParse(tboxAnti.Text, out maxAnti);

            minSup = Math.Min(minSup, 1);
            minSup = Math.Max(minSup, minSupAbs);
            minConf = Math.Min(minConf, 1);
            minConf = Math.Max(minConf, minConfAbs);
            minF = Math.Min(minF, 1);
            minF = Math.Max(minF, minFAbs);
            minS = Math.Min(minS, 1);
            minS = Math.Max(minS, minSAbs);
            maxAnti = Math.Min(maxAnti, 1);
            maxAnti = Math.Min(maxAnti, maxAntiAbs);

            Painter.SetMinValues(minSup, minConf, minS, minF, maxAnti); 
        }

        /// <summary> 
        /// Starts drawing the lasso.
        /// </summary>
        protected void Mouse_Down(object sender, MouseEventArgs e)
        {
            Selection = new GraphicsPath();
            LastPoint.X = e.X;
            LastPoint.Y = e.Y;
            OrigPoint = LastPoint;
            if (sender == pictureBox1) selectMode = 1;
            else if (sender == pictureBox2) selectMode = 2;
            else if (sender == pictureBox3) selectMode = 3;
            else selectMode = 4;
            PaintCharts();
        }
      
        /// <summary> 
        /// Closes the lasso.
        /// </summary>
        protected void Mouse_Up(object sender, MouseEventArgs e)
        {
            int yValue = 0;
            int xValue, CType, lastPareto = 0;
            bool paretoOptimal = true;
            bool clicked;
            bool visible = false;

            if (sender == pictureBox1) CType = 1;
            else if (sender == pictureBox2) CType = 2;
            else if (sender == pictureBox3) CType = 3;
            else CType = 4;

            if (Selection != null)
            {
                Selection.CloseFigure();
                OrigPoint = Point.Empty;
                if (Selection.GetBounds().Width < 2)
                    selectMode = 0;
           
            PaintCharts();
            DGV.RowCount = 0;

            for (int idx = 0; idx < Rules.Count; idx++)
            {
                #region[pareto-optimality check]
                xValue = (int)((Rules[idx].Support * 10 + 1.0) * paddingW);
                switch (CType)
                {
                    case 1:
                        yValue = CHeight - (int)((Rules[idx].Confidence * 10.0 + 1.0) * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].Confidence >= minConf);
                        if (idx == 0 || Rules[idx].Confidence > Rules[lastPareto].Confidence)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 2:
                        yValue = (CHeight - 6 * paddingH) - (int)(Rules[idx].S * 5.0 * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].S >= minS);
                        if (idx == 0 || Rules[idx].S > Rules[lastPareto].S)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 3:
                        yValue = (CHeight - 6 * paddingH) - (int)(Rules[idx].F * 5.0 * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].F >= minF);
                        if (idx == 0 || Rules[idx].F > Rules[lastPareto].F)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;

                    case 4:
                        yValue = CHeight - (int)((Rules[idx].AntiSupport * 10.0 + 1.0) * paddingH);
                        visible = (Rules[idx].Support >= minSup && Rules[idx].AntiSupport <= maxAnti);
                        if (idx == 0 || Rules[idx].AntiSupport < Rules[lastPareto].AntiSupport)
                        {
                            paretoOptimal = true;
                            lastPareto = idx;
                        }
                        else
                            paretoOptimal = false;
                        break;
                }
                #endregion

                clicked = (e.X > xValue - 5 && e.X < xValue + 5 &&
                           e.Y > yValue - 5 && e.Y < yValue + 5);
                paretoOptimal = paretoOptimal || !cboxPareto.Checked;

                if ((Selection.IsVisible(xValue, yValue) || clicked) &&
                    ((visible && paretoOptimal) || cboxInvisible.Checked))
                {
                    AddToGrid(Rules[idx]);
                }
            }
            DGV.AutoResizeColumns();
            }
        }

        /// <summary> 
        /// Draws the lasso.
        /// </summary>
        protected void Mouse_Move(object sender, MouseEventArgs e)
        {
            if (OrigPoint != Point.Empty)
            {
                Point p = new Point(e.X, e.Y);
                Selection.AddLine(LastPoint, p);
                LastPoint.X = e.X;
                LastPoint.Y = e.Y;
                PaintCharts();
            }
        }

        /// <summary>
        /// Paints a single chart depending on the sender.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void pboxPaint(object sender, PaintEventArgs e)
        {
            BRCount = RRCount = 0;
            Graphics CGraphics = e.Graphics;
            int CType;
            double minYValue;

            CalculateMinValues();

            if (sender == pictureBox1) { CType = 1;  minYValue = minConf; }
            else if (sender == pictureBox2) { CType = 2;  minYValue = minS; }
            else if (sender == pictureBox3) { CType = 3; minYValue = minF; }
            else { CType = 4;  minYValue = maxAnti; } 

            CGraphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.Default;
            CGraphics.Clear(Color.White);
            Painter.DrawNet(CGraphics,300,300,CType,Rules[0], 6);
            CGraphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            if (cboxBorder.Checked)
                Painter.DrawBorder(CGraphics, 300, 300, CType, Rules[0]);

            if (cboxSelect.Checked)
            {
                Painter.DrawSelected(CGraphics, 300, 300,
                                    CType, Rules, ref RRCount,ref BRCount, 
                                    cboxInvisible.Checked, Selection, selectMode);
            }
            else if (cboxPareto.Checked && !cboxInvisible.Checked)
            {
                Painter.DrawParetoRule(CGraphics, 300, 300, CType, Rules, 
                                       ref RRCount, ref BRCount);
            }
            else if (cboxPareto.Checked && cboxInvisible.Checked)
            {
                Painter.DrawParetoRuleVisible(CGraphics, 300, 300, CType, 
                                            Rules, ref RRCount,ref BRCount);
            }
            else
            {
                Painter.DrawRule(CGraphics, 300, 300,
                                 CType, Rules,  ref RRCount,
                                 ref BRCount, cboxInvisible.Checked);
            }
            if (selectMode == CType)
                Painter.DrawLasso(CGraphics, Selection);
            Painter.DrawRuleCount(CGraphics, 300, 300,RRCount, BRCount,6);
        }

        /// <summary> 
        /// Forces the chart to repaint itself.
        /// </summary>
        protected void Change(object sender, EventArgs e)
        {
            PaintCharts();
        }

        /// <summary> 
        /// Adds a rule to the datagrid.
        /// </summary>
        /// <param name="r">Rule to be added to the datagrid</param>
        protected void AddToGrid(Rule r)
        {
            int n = DGV.RowCount++;
            DGV[0, n].Value = SetUtil.ParseEvidence(r); ;
            DGV[1, n].Value = SetUtil.ParseHipothesis(r);
            DGV[2, n].Value = String.Format("{0,2:F2}", r.Support);
            DGV[3, n].Value = String.Format("{0,2:F2}", r.Confidence);
            DGV[4, n].Value = String.Format("{0,2:F2}", r.S);
            DGV[5, n].Value = String.Format("{0,2:F2}", r.F);
            DGV[6, n].Value = String.Format("{0,2:F2}", r.AntiSupport);
        }
    }
}