using System;
using System.Collections.Generic;
using System.Text;
using Iesi.Collections;
using System.Windows.Forms;

namespace Association_mining
{
    class RuleComparer : IComparer<Rule>
    {
        public int Compare(Rule r1, Rule r2)
        {
            if (r1.Support < r2.Support)
                return 1;
            else if (r1.Support == r2.Support)
            {
                if (r1.AntiSupport > r2.AntiSupport)
                    return 1;
                else
                    return -1;
            }
            else
                return -1;
        }
    }

    /// <summary>
    /// Rule class. Contains all the nesesary information about a rule like
    /// it's support, confidence, confirmation and others.
    /// </summary>
    public class Rule
    {
        private Set Evidence = new HashedSet(new Item[] {});
        private Set Hipothesis = new HashedSet(new Item[] {});
        private double m_Support;
        private double m_Confidence;
        private double m_AntiSupport;
        private double m_F;
        private double m_S;
        private int jump;
        
        /// <summary>
        /// The "a" value. Number of rules in the database
        /// where E and H.
        /// </summary>
        public long a;
        
        /// <summary>
        /// The "b" value. Number of rules in the database
        /// where ~E andH
        /// </summary>
        public long b;
        
        /// <summary>
        /// The "c" value. Number of rules in the database
        /// where E and ~H
        /// </summary>
        public long c;
        
        /// <summary>
        /// The "d" value. Number of rules in the database
        /// where ~E and ~H
        /// </summary>
        public long d;
        
        /// <summary>
        /// Evidence part of the rule.
        /// </summary>
        internal Set E
        {
            get { return Evidence;  }
        }

        /// <summary>
        /// Hipothesis part of the rule.
        /// </summary>
        internal Set H
        {
            get { return Hipothesis; }
        }

        /// <summary>
        /// Rule support.
        /// </summary>
        public double Support
        {
            get { return m_Support; }
            set { m_Support = value; }
        }

        /// <summary>
        /// Rule confidence.
        /// </summary>
        public double Confidence
        {
            get { return m_Confidence; }
            set { m_Confidence = value; }
        }

        /// <summary>
        /// Rule anti-support.
        /// </summary>
        public double AntiSupport
        {
            get { return m_AntiSupport; }
            set { m_AntiSupport = value; }
        }

        /// <summary>
        /// Confirmation-f of the rule.
        /// </summary>
        public double F
        {
            get { return m_F; }
            set { m_F = value; }
        }


        /// <summary>
        /// Confirmation-s of the rule.
        /// </summary>
        public double S
        {
            get { return m_S; }
            set { m_S = value; }
        }

        /// <summary>
        /// A string version of the evidence part of the rule.
        /// </summary>
        public string premise
        {
            get { return SetUtil.ParseEvidence(this); }
        }

        /// <summary>
        /// String translation of the hipothesis part of the rule.
        /// </summary>
        public string conclusion
        {
            get { return SetUtil.ParseHipothesis(this); }
        }

        /// <summary>
        /// Number of rules with the same support and conclusion. This property is required
        /// to achieve minimal cost while searching for pareto rules.
        /// </summary>
        public int EqualSupportCount
        {
            get { return jump; }
            set { jump = value; }
        }

        /// <summary>
        /// Default constructor.
        /// </summary>
        public Rule() {}

        /// <summary>
        /// Creates a new rule.
        /// </summary>
        /// <param name="e">Evidence</param>
        /// <param name="h">Hipothesis</param>
        /// <param name="u">a value</param>
        /// <param name="w">b value</param>
        /// <param name="x">c balue</param>
        /// <param name="y">d value</param>
        public Rule(Set e, Set h, long u, long w, long x, long y)
        {
            Evidence.AddAll(e);
            Hipothesis.AddAll(h);
            a = u;
            b = w;
            c = x;
            d = y;
            m_S = (double)(a * d - b * c) / (double)((a + c) * (b + d));
            m_F = (double)(a*d - b*c) / (double) (a*d+b*c+2*a*c);
            m_AntiSupport = (double)c / (double)(a + b + c + d);
            m_Confidence = (double)a / (double)(a + c);
            m_Support = (double)a / (double)(a + b + c + d);
            jump = 1;
        }

        /// <summary>
        /// Creates a new rule.
        /// </summary>
        /// <param name="e">Evidence</param>
        /// <param name="h">Hipothesis</param>
        /// <param name="u">a value</param>
        /// <param name="w">b value</param>
        /// <param name="x">c balue</param>
        /// <param name="y">d value</param>
        /// <param name="equalCount">Number of rules with the same support and conclusion</param>
        public Rule(Set e, Set h, long u, long w, long x, long y, int equalCount)
        {
            Evidence.AddAll(e);
            Hipothesis.AddAll(h);
            a = u;
            b = w;
            c = x;
            d = y;
            m_S = (double)(a * d - b * c) / (double)((a + c) * (b + d));
            m_F = (double)(a * d - b * c) / (double)(a * d + b * c + 2 * a * c);
            m_AntiSupport = (double)c / (double)(a + b + c + d);
            m_Confidence = (double)a / (double)(a + c);
            m_Support = (double)a / (double)(a + b + c + d);
            jump = equalCount;
        }

        /// <summary>
        /// Transforms a list of unordered rules into a rule structure where rules
        /// with the same xconclusion are on a seperate list sorted descending by 
        /// support and ascending by anti-support. Each rule in the structure knows
        /// the next cell with a different support.
        /// </summary>
        /// <param name="ruleList">List of all the rules</param>
        /// <param name="ConclusionList">List of distinct conclusions</param>
        /// <returns>A rule structure</returns>
        public static List<List<Rule>> Transform(SortableBindingList<Rule> ruleList, List<Set> ConclusionList)
        {
            int conclusionCount = ConclusionList.Count;
            List<List<Rule>> ruleStructure = new List<List<Rule>>(conclusionCount);

            for (int i = 0; i < conclusionCount; i++)
                ruleStructure.Add(new List<Rule>());
            
            foreach (Rule r in ruleList)
            {
                for (int i = 0; i < conclusionCount; i++)
                {
                    if (ConclusionList[i].Equals(r.H))
                    {
                        ruleStructure[i].Add(r);
                        break;
                    }
                }
            }
            SortStructure(ruleStructure);
            ModifyEqualCount(ruleStructure);
            return ruleStructure;
        }

        /// <summary>
        /// Modifies the equal count values in the structure.
        /// </summary>
        /// <param name="ruleStructure">Rule structure</param>
        private static void ModifyEqualCount(List<List<Rule>> ruleStructure)
        {    
            foreach (List<Rule> rulelist in ruleStructure)
            {
                int previousRuleIdx = 0;
                int equalCount = 1;
                for (int i = 1; i < rulelist.Count; i++)
                {
                    if (rulelist[previousRuleIdx].Support == rulelist[i].Support &&
                        rulelist[previousRuleIdx].AntiSupport != rulelist[i].AntiSupport)
                    {
                        equalCount++;
                    }
                    else
                    {
                        rulelist[previousRuleIdx].EqualSupportCount = equalCount;
                        previousRuleIdx = i;
                        equalCount = 1;
                    }

                }
                rulelist[previousRuleIdx].EqualSupportCount = equalCount;
            }
        }

        /// <summary>
        /// Sorts the rule structure descending by 
        /// support and ascending by anti-support.
        /// </summary>
        /// <param name="structure">Rule structure</param>
        public static void SortStructure(List<List<Rule>> structure)
        {
            foreach(List<Rule> list in structure)
                QuickSort(list, 0, list.Count-1);
        }

        /// <summary>
        /// Quicksort method used to sort rules.
        /// </summary>
        /// <param name="list">Rule list</param>
        /// <param name="leftBoundary">Left boundary of the list to sort</param>
        /// <param name="rightBoundary">Right boundary of the list to sort</param>
        protected static void QuickSort(List<Rule> list, int leftBoundary, int rightBoundary)
        {
            if (leftBoundary < rightBoundary)
            {
                int pivotIndex = GetPivot(list, leftBoundary, rightBoundary);
                QuickSort(list, leftBoundary, pivotIndex - 1);
                QuickSort(list, pivotIndex + 1, rightBoundary);
            }
        }
 
        /// <summary>
        /// Used to find the pivot point for quicksort
        /// </summary>
        /// <param name="list">Rule list</param>
        /// <param name="leftBoundary">Left boundary of the list to sort</param>
        /// <param name="rightBoundary">Right boundary of the list to sort</param>
        /// <returns>Index of the element to become the pivot point for quicksort</returns>
        protected static int GetPivot(List<Rule> list, int leftBoundary, int rightBoundary)
        {
            RuleComparer comparer = new RuleComparer();
            int middle = (leftBoundary + rightBoundary) / 2;

            if (comparer.Compare(list[leftBoundary], list[rightBoundary]) < 0)
            {
                Swap(list, leftBoundary, rightBoundary);
            }

            if (comparer.Compare(list[middle], list[rightBoundary]) < 0)
            {
                Swap(list, middle, rightBoundary);
            }

            if (comparer.Compare(list[leftBoundary], list[middle]) > 0)
            {
                Swap(list, leftBoundary, middle);
            }

            int pivotIndex = leftBoundary;
            Rule pivot = list[pivotIndex];

            for (int i = leftBoundary + 1; i <= rightBoundary; i++)
            {
                if (comparer.Compare(list[i], pivot) < 0)
                {
                    pivotIndex++;
                    Swap(list, pivotIndex, i);
                }
            }

            Swap(list, leftBoundary, pivotIndex);

            return pivotIndex;
        }

        /// <summary>
        /// Swaps two rules.
        /// </summary>
        /// <param name="list">Rule list</param>
        /// <param name="pos1">First rule</param>
        /// <param name="pos2">Second rule</param>
        protected static void Swap(List<Rule> list, int pos1, int pos2)
        {
            Rule tmp = list[pos1];
            list[pos1] = list[pos2];
            list[pos2] = tmp;
        }
    }
}
