using System;
using System.Collections.Generic;
using System.Text;
using Iesi.Collections;
using System.Collections;

namespace Association_mining
{
    /// <summary>
    /// A class representing a single item. As an item we understand
    /// a unique pair of an attribute type and name.
    /// </summary>
    public class Item
    {
        /// <summary>
        /// Attribute type.
        /// </summary>
        public string Type;

        /// <summary>
        /// Attribute name.
        /// </summary>
        public string Name;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public Item() { }

        /// <summary>
        /// Creates a new item.
        /// </summary>
        /// <param name="t">Type</param>
        /// <param name="n">Name</param>
        public Item(string t, string n)
        {
            Type = t;
            Name = n;
        }

        /// <summary>
        /// Overridden equals method allows to compare items.
        /// Items are said to be equal if they have identical
        /// <c>Type</c> and <c>Name</c> values.
        /// </summary>
        /// <param name="obj"></param>
        /// <returns>True if items are identical and false 
        /// if they are not identical</returns>
        public override bool Equals(object obj)
        {
            if(obj == null) return false;
            if (this.GetType() != obj.GetType()) return false;
            Item otherObj = (Item)obj;
            if (!Type.Equals(otherObj.Type)) return false;
            if (!Name.Equals(otherObj.Name)) return false;
            return true;
        }

        /// <summary>
        /// Returns a unique hash value based on the item type and name.
        /// </summary>
        /// <returns>A unique value representing an item</returns>
        public override int GetHashCode()
        {
            return (Type+Name).GetHashCode();
        }
    }

    /// <summary>
    /// Class representing a frequent set. That is a set
    /// with a known support and universe count (size of the database)
    /// </summary>
    public class FrequentSet
    {
        /// <summary>
        /// Number of items in the frequent set.
        /// </summary>
        public int Count
        {
            get { return set.Count; }
        }

        /// <summary>
        /// The set.
        /// </summary>
        public Set set;

        /// <summary>
        /// Frequent set support.
        /// </summary>
        public long Support;

        /// <summary>
        /// Universe count.
        /// </summary>
        public int UniverseCount;

        /// <summary>
        /// Default constructor.
        /// </summary>
        public FrequentSet() { }

        /// <summary>
        /// Creates a new frequent set based on a given set,
        /// support value and database size.
        /// </summary>
        /// <param name="s"></param>
        /// <param name="sup"></param>
        /// <param name="u"></param>
        public FrequentSet(Set s, long sup, int u)
        {
            set = s;
            Support = sup;
            UniverseCount = u;
        }

        /// <summary>
        /// Overridden equals method allows to compare frequent sets.
        /// Frequent sets are said tio be identical if they have the same
        /// set of items, support and universe count.
        /// </summary>
        /// <param name="obj"></param>
        /// <returns>True if frequent sets are identical</returns>
        public override bool Equals(object obj)
        {
            if (obj == null) return false;
            if (this.GetType() != obj.GetType()) return false;
            FrequentSet otherObj = (FrequentSet)obj;
            if (!set.Equals(otherObj.set)) return false;
            if (!Support.Equals(otherObj.Support)) return false;
            if (!UniverseCount.Equals(otherObj.UniverseCount)) return false;
            return true;
        }

        /// <summary>
        /// Returns a unique hash value for the frequent set.
        /// </summary>
        /// <returns>Unique hash value for the frequent set</returns>
        public override int GetHashCode()
        {
            return (set.GetHashCode() + (Support+UniverseCount).GetHashCode());
        }
    }

    /// <summary>
    /// A static class providing basic functions that operate on sets.
    /// This includes rule generation, creation of subsets, parsing sets to strings etc.
    /// </summary>
    public static class SetUtil
    {
        /// <summary>
        /// Generates a list of subsets with a count htat is one less in 
        /// comparison to the parameter set.
        /// </summary>
        /// <param name="set">Set to create a subset from</param>
        /// <returns>A list of subsets that have one item less
        /// than the primary set</returns>
        public static List<Set> SubsetMinusOne(Set set)
        {
            List<Set> SubsetList = new List<Set>();
            Set subset = new HashedSet();
            foreach (Item i in set)
            {
                subset = set - new HashedSet(new Item[] { i });
                SubsetList.Add(subset);
            }
            return SubsetList;
        }

        /// <summary>
        /// Checks if all the items in the given set are of different types.
        /// </summary>
        /// <param name="s">Primary set</param>
        /// <returns>True if none of the items of the items in the
        /// set are of the same attribute type</returns>
        public static bool HasDistinctTypes(Set s)
        {
            Dictionary<string, int> hash = new Dictionary<string, int>();
            foreach (Item i in s)
            {
                if (!hash.ContainsKey(i.Type))
                    hash.Add(i.Type, 0);
                else
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Generates candidate sets from a given list of sets.
        /// </summary>
        /// <param name="setList">List of sets</param>
        /// <returns>Sets that will possibly be frequent sets</returns>
        public static List<Set> GenerateCandidateSets(List<Set> setList)
        {
            List<Set> LargerSets = new List<Set>();
            List<Set> CandidateSets = new List<Set>();
            List<Set> SubsetList = new List<Set>();

            Set union;
            foreach (Set e1 in setList)
            {
                foreach (Set e2 in setList)
                {
                    union = e1 | e2;
                    if (union.Count == e1.Count + 1 && HasDistinctTypes(union) && !LargerSets.Contains(union))
                        LargerSets.Add(union);
                }
            }
            //set prunning
            foreach (Set s in LargerSets)
            {
                bool hasAllSubsets = true;
                SubsetList = SubsetMinusOne(s);
                foreach (Set subset in SubsetList)
                {
                    if (!setList.Contains(subset))
                    {
                        hasAllSubsets = false;
                        break;
                    }
                }
                if (hasAllSubsets)
                    CandidateSets.Add(s);
            }
            return CandidateSets;
        }

        private static bool ContainsType(Set union, string decClass)
        {
            bool contains = false;
            foreach (Item i in union)
            {
                if (i.Type == decClass)
                    contains = true;
            }
            return contains;
        }

        /// <summary>
        /// Generates a Johnson code transition vector used to
        /// find all subsets.
        /// </summary>
        /// <param name="n">The number of items in the set to look for subsets</param>
        /// <returns>An array of transition positions</returns>
        public static int[] GenerateTransitionVector(int n) {

		    int i,k,Twsk,j;
		    bool flag;
		    int[] v = new int[n];
		    int warunekZak;

		    i = 1; 
		    Twsk = 0;
		    flag = true;
		    warunekZak = (int) (Math.Pow(2,n)-1);
		    
		    int[] Tout = new int[warunekZak];

		    for (j=0; j<n; j++)
		    	v[j] = j;

		    while (flag) {
		      if (v[i]>0) {
		        Tout[Twsk] = v[i];
		        Twsk++;
		        v[i] = 0;
		        for (k=i-1; k>=1; k--)
		        	v[k] = k;
		        i = 1;
		      } else
		        i++;
		      
		      if (Twsk ==((warunekZak) / 2 )) 
		    	  flag = false;
		    }
		    Tout[Twsk] = n;

		    j = 1;
		    for (i=Twsk+1; i<warunekZak; i++) {
		      Tout[i] = Tout[Twsk-j];
		      j++;
		    }

		    return Tout;
	  }

        /// <summary>
        /// Chaecks the support of a given (frequent) set.
        /// </summary>
        /// <param name="s">The checked set</param>
        /// <param name="fsList">Frequent set list</param>
        /// <param name="data">Database</param>
        /// <returns>The support value of the set</returns>
        public static long CheckSupport(Set s, List<FrequentSet> fsList, Database data)
        {
            if (s.Count == 1) 
            {
                foreach(Item i in s)
                    return data.Template.AttributeSupport(data.Translator[i.Type+i.Name]);
            }
            else
            {
                foreach (FrequentSet fs in fsList)
                {
                    if (fs.set.Equals(s))
                        return fs.Support;
                }
            }
            //tu powinien byc wyjatek!
            return long.MaxValue;
        }

        /// <summary>
        /// Generates a list of subsets of the given set.
        /// </summary>
        /// <param name="s">Primary set</param>
        /// <returns>A list of subsets</returns>
        public static List<Set> Subsets(Set s)
        {
            List<Set> list = new List<Set>();
            FPNode[] ItemArray = new FPNode[s.Count];
            Set Subset = new HashedSet(new FPNode[] { });
            int n = 0;

            if(s.Count == 1)
            {
                list.Add(new HashedSet(s));
                return list;
            }
            if(s.Count == 2)
            {
                list.Add(new HashedSet(s));
                foreach (FPNode fpn in s)
                    list.Add(new HashedSet(new FPNode[] { fpn }));
                return list;
            }

            foreach (FPNode i in s)
            {
                ItemArray[n] = i;
                n++;
            }

            int[] TransitionVector = GenerateTransitionVector(s.Count);
            for (int i = 0; i < TransitionVector.Length; i++)
            {
                if (!Subset.Add(ItemArray[TransitionVector[i] - 1]))
                    Subset.Remove(ItemArray[TransitionVector[i] - 1]);

                if (Subset.Count > 0)
                    list.Add(new HashedSet(Subset));
            }
            return list;
        }

        /// <summary>
        /// Generates rules from a list of frequent sets.
        /// </summary>
        /// <param name="MainForm">Main program form</param>
        /// <param name="data">Database</param>
        /// <param name="Conclusions">Conclusion list</param>
        /// <param name="frequentsets">List of frequent sets</param>
        /// <param name="maxanti">Maximal anti-support</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <returns>A list of rules</returns>
        public static SortableBindingList<Rule> GenerateRules(frmMain MainForm, Database data, List<Set> Conclusions,
                List<FrequentSet> frequentsets, double minconf, double mins, double minf, double maxanti)
        {
            int previousFsCount = 2;
            int[] TransitionVector = new int[] { };
            long evidenceSupport;
            Rule r = new Rule();
            Set FSMinusS = new HashedSet(new Item[] {});
            SortableBindingList<Rule> RuleList = new SortableBindingList<Rule>();

            foreach (FrequentSet fs in frequentsets)
            {
                int fsCount = fs.Count;

                #region [zbior dwuelementowy]
                if (fsCount == 2)
                {
                    Item[] items = new Item[2];
                    int n = 0;
                    foreach (Item i in fs.set)
                    {
                        items[n] = i;
                        n++;
                    }
                    //regula 1
                    long a, b, c;
                    r = new Rule(new HashedSet(new Item[] { items[0] }),               //E
                                      new HashedSet(new Item[] { items[1] }),               //H
                                      a = fs.Support,                                         //a
                                      b = data.Template.AttributeSupport(data.Translator[items[1].Type + items[1].Name])
                                      - fs.Support,                                       //b
                                      c = data.Template.AttributeSupport(data.Translator[items[0].Type + items[0].Name])
                                      - fs.Support,                                       //c
                                      (long)fs.UniverseCount - (a + b + c));
                    if (r.Confidence >= minconf && r.S >= mins &&
                                r.F >= minf && r.AntiSupport <= maxanti)
                    {
                        RuleList.Add(r);
                        if (!Conclusions.Contains(r.H))
                            Conclusions.Add(r.H);
                    }

                    //regula 2
                    r = new Rule(new HashedSet(new Item[] { items[1] }),               //E
                                      new HashedSet(new Item[] { items[0] }),               //H
                                      a = fs.Support,                                       //a
                                      b = data.Template.AttributeSupport(data.Translator[items[0].Type + items[0].Name])
                                      - fs.Support,                                         //b
                                      c = data.Template.AttributeSupport(data.Translator[items[1].Type + items[1].Name])
                                      - fs.Support,                                         //c
                                      (long)fs.UniverseCount - (a + b + c));               //d
                    if (r.Confidence >= minconf && r.S >= mins &&
                         r.F >= minf && r.AntiSupport <= maxanti)
                    {
                        RuleList.Add(r);
                        if (!Conclusions.Contains(r.H))
                            Conclusions.Add(r.H);
                    }
                }
                #endregion
                #region [zbior wiecej niz dwu- elementowy]
                else
                {
                    //generujemy tablice przejsc tylko gdy zmienila sie licznosc zbioru
                    if (fsCount != previousFsCount)
                        TransitionVector = GenerateTransitionVector(fsCount);

                    Item [] ItemArray = new Item[fsCount];
                    Set Subset = new HashedSet(new Item[] { });

                    //tworzymy tablice elementow tworzacych zbior
                    int n = 0;
                    foreach (Item i in fs.set)
                    {
                        ItemArray[n] = i;
                        n++;
                    }

                    for (int i = 0; i < TransitionVector.Length; i++)
                    {
                        //tworzenie nowego podzbioru zgodnie z tablica przejsc
                        //jak sie nie udalo dodac to odejmujemy...
                        if(!Subset.Add(ItemArray[TransitionVector[i]-1]))
                            Subset.Remove(ItemArray[TransitionVector[i]-1]);

                        if (Subset.Count<fsCount && Subset.Count >0)
                        {
                            long a, b, c;
                            evidenceSupport = CheckSupport(Subset, frequentsets, data);
                            r =      new Rule(Subset,                                                       //E
                                              FSMinusS = fs.set - Subset,                                   //H
                                              a = fs.Support,                                               //a
                                              b = CheckSupport(FSMinusS,frequentsets,data) - fs.Support,     //b
                                              c = evidenceSupport- fs.Support,                              //c
                                              (long)fs.UniverseCount - (a + b + c));                       //d

                            if (r.Confidence >= minconf && r.S >= mins &&
                                r.F >= minf && r.AntiSupport <= maxanti)
                            {
                                RuleList.Add(r);
                                if (!Conclusions.Contains(r.H))
                                    Conclusions.Add(r.H);
                            }
                        }

                    }
                }
                #endregion
                MainForm.SetRuleCount(RuleList.Count);
            }
            return RuleList;
        }

        /// <summary>
        /// Generates rules from a list of frequent sets.
        /// </summary>
        /// <param name="MainForm">Main program form</param>
        /// <param name="data">Database</param>
        /// <param name="Conclusions">Conclusion list</param>
        /// <param name="frequentsets">List of frequent sets</param>
        /// <param name="maxanti">Maximal anti-support</param>
        /// <param name="minconf">Minimal confidence</param>
        /// <param name="minf">Minimal confirmation-f</param>
        /// <param name="mins">Minimal confirmation-s</param>
        /// <returns>A list of rules</returns>
        public static SortableBindingList<Rule> GenerateRules(frmMain MainForm, Database data, List<Set> Conclusions,
                List<FrequentSet> frequentsets, double minconf, double mins, double minf, double maxanti, string decClass)
        {
            int previousFsCount = 2;
            int[] TransitionVector = new int[] { };
            long evidenceSupport;
            Rule r = new Rule();
            Set FSMinusS = new HashedSet(new Item[] { });
            SortableBindingList<Rule> RuleList = new SortableBindingList<Rule>();

            foreach (FrequentSet fs in frequentsets)
            {
                int fsCount = fs.Count;

                #region [zbior dwuelementowy]
                if (fsCount == 2)
                {
                    Item[] items = new Item[2];
                    int n = 0;
                    foreach (Item i in fs.set)
                    {
                        items[n] = i;
                        n++;
                    }
                    long a, b, c;
                    if (items[1].Type == decClass)
                    {
                        
                        r = new Rule(new HashedSet(new Item[] { items[0] }),               //E
                                          new HashedSet(new Item[] { items[1] }),               //H
                                          a = fs.Support,                                         //a
                                          b = data.Template.AttributeSupport(data.Translator[items[1].Type + items[1].Name])
                                          - fs.Support,                                       //b
                                          c = data.Template.AttributeSupport(data.Translator[items[0].Type + items[0].Name])
                                          - fs.Support,                                       //c
                                          (long)fs.UniverseCount - (a + b + c));
                        if (r.Confidence >= minconf && r.S >= mins &&
                                    r.F >= minf && r.AntiSupport <= maxanti)
                        {
                            RuleList.Add(r);
                            if (!Conclusions.Contains(r.H))
                                Conclusions.Add(r.H);
                        }
                    }
                    if (items[0].Type == decClass)
                    {
                        //regula 2
                        r = new Rule(new HashedSet(new Item[] { items[1] }),               //E
                                          new HashedSet(new Item[] { items[0] }),               //H
                                          a = fs.Support,                                       //a
                                          b = data.Template.AttributeSupport(data.Translator[items[0].Type + items[0].Name])
                                          - fs.Support,                                         //b
                                          c = data.Template.AttributeSupport(data.Translator[items[1].Type + items[1].Name])
                                          - fs.Support,                                         //c
                                          (long)fs.UniverseCount - (a + b + c));               //d
                        if (r.Confidence >= minconf && r.S >= mins &&
                             r.F >= minf && r.AntiSupport <= maxanti)
                        {
                            RuleList.Add(r);
                            if (!Conclusions.Contains(r.H))
                                Conclusions.Add(r.H);
                        }
                    }
                }
                #endregion
                #region [zbior wiecej niz dwu- elementowy]
                else
                {
                    Item[] ItemArray = new Item[fsCount];
                    Set Subset = new HashedSet(new Item[] { });

                    //tworzymy tablice elementow tworzacych zbior
                    foreach (Item i in fs.set)
                    {
                        if (i.Type != decClass)
                            Subset.Add(i);
                    }

                    if (Subset.Count < fsCount && Subset.Count > 0)
                    {
                        long a, b, c;
                        evidenceSupport = CheckSupport(Subset, frequentsets, data);
                        r = new Rule(Subset,                                                       //E
                                          FSMinusS = fs.set - Subset,                                   //H
                                          a = fs.Support,                                               //a
                                          b = CheckSupport(FSMinusS, frequentsets, data) - fs.Support,     //b
                                          c = evidenceSupport - fs.Support,                              //c
                                          (long)fs.UniverseCount - (a + b + c));                       //d

                        if (r.Confidence >= minconf && r.S >= mins &&
                            r.F >= minf && r.AntiSupport <= maxanti)
                        {
                            RuleList.Add(r);
                            if (!Conclusions.Contains(r.H))
                                Conclusions.Add(r.H);
                        }
                    }
                }
                #endregion
                MainForm.SetRuleCount(RuleList.Count);
            }
            return RuleList;
        }

        /// <summary>
        /// Transform the rules evidence from set form to a string.
        /// </summary>
        /// <param name="r">Rule which evidence will be parsed</param>
        /// <returns>A string representing the rules evidence</returns>
        public static string ParseEvidence(Rule r)
        {
            string text="";

            foreach (Item it in r.E)
                text += it.Type + " is " + it.Name + " and ";
            text = text.Substring(0, text.Length - 5);
            return text;
        }

        /// <summary>
        /// Transform from set form to a string.
        /// </summary>
        /// <returns>A string representing the set</returns>
        public static string ParseSet(Set s)
        {
            string text = "";

            foreach (Item it in s)
                text += it.Type + " is " + it.Name + " and ";
            text = text.Substring(0, text.Length - 5);
            return text;
        }

        /// <summary>
        /// Transform the rules hipothesis from set form to a string.
        /// </summary>
        /// <param name="r">Rule which hipothesis will be parsed</param>
        /// <returns>A string representing the rules hipothesis</returns>
        public static string ParseHipothesis(Rule r)
        {
            string text="";

            foreach (Item it in r.H)
                text += it.Type + " is " + it.Name + " and ";
            text = text.Substring(0, text.Length - 5);
            return text;
        }
    }
}
